import warnings

import torch.nn as nn

from neuralhydrology.modelzoo.arlstm import ARLSTM
from neuralhydrology.modelzoo.cudalstm import CudaLSTM
from neuralhydrology.modelzoo.customlstm import CustomLSTM
from neuralhydrology.modelzoo.ealstm import EALSTM
from neuralhydrology.modelzoo.embcudalstm import EmbCudaLSTM
from neuralhydrology.modelzoo.handoff_forecast_lstm import HandoffForecastLSTM
from neuralhydrology.modelzoo.hybridmodel import HybridModel
from neuralhydrology.modelzoo.gru import GRU
from neuralhydrology.modelzoo.mclstm import MCLSTM
from neuralhydrology.modelzoo.mtslstm import MTSLSTM
from neuralhydrology.modelzoo.multihead_forecast_lstm import MultiHeadForecastLSTM
from neuralhydrology.modelzoo.odelstm import ODELSTM
from neuralhydrology.modelzoo.sequential_forecast_lstm import SequentialForecastLSTM
from neuralhydrology.modelzoo.stacked_forecast_lstm import StackedForecastLSTM
from neuralhydrology.modelzoo.transformer import Transformer
from neuralhydrology.utils.config import Config

SINGLE_FREQ_MODELS = [
    "cudalstm", 
    "ealstm", 
    "customlstm", 
    "embcudalstm", 
    "gru", 
    "transformer", 
    "mclstm", 
    "arlstm",
    "handoff_forecast_lstm",
    "sequential_forecast_lstm",
    "multihead_forecast_lstm",
    "stacked_forecast_lstm"
]
AUTOREGRESSIVE_MODELS = ['arlstm']


def get_model(cfg: Config) -> nn.Module:
    """Get model object, depending on the run configuration.
    
    Parameters
    ----------
    cfg : Config
        The run configuration.

    Returns
    -------
    nn.Module
        A new model instance of the type specified in the config.
    """
    if cfg.model.lower() in SINGLE_FREQ_MODELS and len(cfg.use_frequencies) > 1:
        raise ValueError(f"Model {cfg.model} does not support multiple frequencies.")

    if cfg.model.lower() not in AUTOREGRESSIVE_MODELS and cfg.autoregressive_inputs:
        raise ValueError(f"Model {cfg.model} does not support autoregression.")

    if cfg.model.lower() != "mclstm" and cfg.mass_inputs:
        raise ValueError(f"The use of 'mass_inputs' with {cfg.model} is not supported.")

    if cfg.model.lower() == "arlstm":
        model = ARLSTM(cfg=cfg)
    elif cfg.model.lower() == "cudalstm":
        model = CudaLSTM(cfg=cfg)
    elif cfg.model.lower() == "ealstm":
        model = EALSTM(cfg=cfg)
    elif cfg.model.lower() == "customlstm":
        model = CustomLSTM(cfg=cfg)
    elif cfg.model.lower() == "lstm":
        warnings.warn(
            "The `LSTM` class has been renamed to `CustomLSTM`. Support for `LSTM` will we dropped in the future.",
            FutureWarning)
        model = CustomLSTM(cfg=cfg)
    elif cfg.model.lower() == "gru":
        model = GRU(cfg=cfg)
    elif cfg.model.lower() == "embcudalstm":
        model = EmbCudaLSTM(cfg=cfg)
    elif cfg.model.lower() == "mtslstm":
        model = MTSLSTM(cfg=cfg)
    elif cfg.model.lower() == "odelstm":
        model = ODELSTM(cfg=cfg)
    elif cfg.model.lower() == "mclstm":
        model = MCLSTM(cfg=cfg)
    elif cfg.model.lower() == "transformer":
        model = Transformer(cfg=cfg)
    elif cfg.model.lower() == "handoff_forecast_lstm":
        model = HandoffForecastLSTM(cfg=cfg)
    elif cfg.model.lower() == "multihead_forecast_lstm":
        model = MultiHeadForecastLSTM(cfg=cfg)
    elif cfg.model.lower() == "sequential_forecast_lstm":
        model = SequentialForecastLSTM(cfg=cfg)
    elif cfg.model.lower() == "stacked_forecast_lstm":
        model = StackedForecastLSTM(cfg=cfg)
    elif cfg.model.lower() == "hybrid_model":
        model = HybridModel(cfg=cfg)
    else:
        raise NotImplementedError(f"{cfg.model} not implemented or not linked in `get_model()`")

    return model



from typing import Callable, Dict, List, Union

import numpy as np
import pandas as pd
import torch
import xarray
from numba import njit
from torch.distributions import Categorical

from neuralhydrology.utils.config import Config


def sample_pointpredictions(model: 'BaseModel', data: Dict[str, torch.Tensor], n_samples: int,
                            scaler: Dict[str, Union[pd.Series, xarray.Dataset]]) -> Dict[str, torch.Tensor]:
    """Point prediction samplers for the different uncertainty estimation approaches.
    
    This function provides different point sampling functions for the different uncertainty estimation approaches 
    (i.e. Gaussian Mixture Models (GMM), Countable Mixtures of Asymmetric Laplacians (CMAL), Uncountable Mixtures of 
    Asymmetric Laplacians (UMAL), and Monte-Carlo Dropout (MCD); note: MCD can be combined with the others, by setting 
    `mc_dropout` to `True` in the configuration file). 
    
    There are also options to handle negative point prediction samples that arise while sampling from the uncertainty 
    estimates. This functionality currently supports (a) 'clip' for directly clipping values at zero and 
    (b) 'truncate' for resampling values that are below zero. 
    
    Parameters
    ----------
    model : BaseModel
        The neuralhydrology model from which to sample from.
    data : Dict[str, torch.Tensor]
        Dictionary, containing input features as key-value pairs.
    n_samples : int
        The number of point prediction samples that should be created.
    scaler : Dict[str, Union[pd.Series, xarray.Dataset]]
        Scaler of the run.

    Returns
    -------
    Dict[str, torch.Tensor]
        Dictionary, containing the sampled model outputs for the `predict_last_n` (config argument) time steps of 
        each frequency.
    """

    if model.cfg.head.lower() == "gmm":
        samples = sample_gmm(model, data, n_samples, scaler)
    elif model.cfg.head.lower() == "cmal":
        samples = sample_cmal(model, data, n_samples, scaler)
    elif model.cfg.head.lower() == "umal":
        samples = sample_umal(model, data, n_samples, scaler)
    elif model.cfg.head.lower() == "regression":
        samples = sample_mcd(model, data, n_samples, scaler)  # regression head assumes mcd
    else:
        raise NotImplementedError(f"Sampling mode not supported for head {model.cfg.head.lower()}!")

    return samples


def _subset_target(parameter: Dict[str, torch.Tensor], n_target: int, steps: int) -> Dict[str, torch.Tensor]:
    # determine which output neurons correspond to the n_target target variable
    start = n_target * steps
    end = (n_target + 1) * steps
    parameter_sub = parameter[:, :, start:end]

    return parameter_sub


def _handle_negative_values(cfg: Config, values: torch.Tensor, sample_values: Callable,
                            scaler: Dict[str, Union[pd.Series, xarray.Dataset]], nth_target: int) -> torch.Tensor:
    """Handle negative samples that arise while sampling from the uncertainty estimates.

    Currently supports (a) 'clip' for directly clipping values at zero and (b) 'truncate' for resampling values 
    that are below zero. 

    Parameters
    ----------
    cfg : Config
        The run configuration.
    values : torch.Tensor
        Tensor with the sampled values.
    sample_values : Callable
        Sampling function to allow for repeated sampling in the case of truncation-handling. 
    scaler : Dict[str, Union[pd.Series, xarray.Dataset]]
        Scaler of the run.
    nth_target : int
        Index of the sampled target variable in cfg.target_variables.

    Returns
    -------
    torch.Tensor
        Bound values according to user specifications.  
    """
    center = scaler["xarray_feature_center"][cfg.target_variables[nth_target]]
    scale = scaler["xarray_feature_scale"][cfg.target_variables[nth_target]]
    normalized_zero = -torch.from_numpy((center / scale).values).to(values)
    if cfg.negative_sample_handling.lower() == 'clip':
        values = torch.clamp(values, min=normalized_zero)
    elif cfg.negative_sample_handling.lower() == 'truncate':
        values_smaller_zero = values < normalized_zero
        try_count = 0
        while torch.any(values_smaller_zero.flatten()):
            values[values_smaller_zero] = sample_values(values_smaller_zero)
            values_smaller_zero = values < normalized_zero
            try_count += 1
            if try_count >= cfg.negative_sample_max_retries:
                break
    elif cfg.negative_sample_handling is None or cfg.negative_sample_handling.lower() == 'none':
        pass
    else:
        raise NotImplementedError(
            f"The option {cfg.negative_sample_handling} is not supported for handling negative samples!")

    return values


def _sample_gaussian_mixtures(ids: List[int], m_sub: torch.Tensor, s_sub: torch.Tensor,
                              p_sub: torch.Tensor) -> torch.Tensor:
    # unbound sampling:
    categorical = Categorical(p_sub)
    pis = categorical.sample().data
    mask_gmm = torch.zeros(p_sub.shape, dtype=torch.bool) \
        .to(device=p_sub.device) \
        .scatter_(2, pis.unsqueeze(2), True)

    # The ids are used for location-specific resampling for 'truncation' in '_handle_negative_values'
    values = s_sub \
        .data.new(s_sub[ids][mask_gmm[ids]].shape[0]) \
        .normal_() \
        .flatten() \
        .mul(s_sub[ids][mask_gmm[ids]]) \
        .add(m_sub[ids][mask_gmm[ids]])
    return values


def _sample_asymmetric_laplacians(ids: List[int], m_sub: torch.Tensor, b_sub: torch.Tensor,
                                  t_sub: torch.Tensor) -> torch.Tensor:
    # The ids are used for location-specific resampling for 'truncation' in '_handle_negative_values'
    prob = torch.FloatTensor(m_sub[ids].shape) \
        .uniform_(0, 1) \
        .to(m_sub.device)  # sample uniformly between zero and 1
    values = torch.where(
        prob < t_sub[ids],  # needs to be in accordance with the loss
        m_sub[ids] + ((b_sub[ids] * torch.log(prob / t_sub[ids])) / (1 - t_sub[ids])),
        m_sub[ids] - ((b_sub[ids] * torch.log((1 - prob) / (1 - t_sub[ids]))) / t_sub[ids]))
    return values.flatten()


class _SamplingSetup():

    def __init__(self, model: 'BaseModel', data: Dict[str, torch.Tensor], head: str):
        # make model checks:
        cfg = model.cfg
        if not cfg.head.lower() == head.lower():
            raise NotImplementedError(f"{head} sampling not supported for the {cfg.head} head!")

        dropout_modules = [model.dropout.p]

        # Certain models don't have embedding_net(s)
        implied_statics_embedding, implied_dynamics_embedding = None, None
        if hasattr(model, 'forecast_embedding_net'):
            implied_forecast_statics_embedding = model.forecast_embedding_net.statics_embedding_p_dropout
            implied_forecast_dynamics_embedding = model.forecast_embedding_net.dynamics_embedding_p_dropout
            dropout_modules += [implied_forecast_statics_embedding, implied_forecast_dynamics_embedding]
        if hasattr(model, 'hindcast_embedding_net'):
            implied_hindcast_statics_embedding = model.hindcast_embedding_net.statics_embedding_p_dropout
            implied_hindcast_dynamics_embedding = model.hindcast_embedding_net.dynamics_embedding_p_dropout
            dropout_modules += [implied_hindcast_statics_embedding, implied_hindcast_dynamics_embedding]
        if hasattr(model, 'embedding_net'):
            implied_statics_embedding = model.embedding_net.statics_embedding_p_dropout
            implied_dynamics_embedding = model.embedding_net.dynamics_embedding_p_dropout
            dropout_modules += [implied_statics_embedding, implied_dynamics_embedding]
        # account for transformer
        implied_transformer_dropout = None
        if cfg.model.lower() == 'transfomer':
            implied_transformer_dropout = cfg.transformer_dropout
            dropout_modules.append(implied_transformer_dropout)

        max_implied_dropout = max(dropout_modules)
        # check lower bound dropout:
        if cfg.mc_dropout and max_implied_dropout <= 0.0:
            raise RuntimeError(f"""{cfg.model} with `mc_dropout` activated requires a dropout rate larger than 0.0
                               The current implied dropout-rates are:
                                  - model: {cfg.output_dropout}
                                  - statics_embedding: {implied_statics_embedding}
                                  - dynamics_embedding: {implied_dynamics_embedding}
                                  - statics_forecast_embedding: {implied_forecast_statics_embedding}
                                  - dynamics_forecast_embedding: {implied_forecast_dynamics_embedding}
                                  - statics_hindcast_embedding: {implied_hindcast_statics_embedding}
                                  - dynamics_hindcast_embedding: {implied_hindcast_dynamics_embedding}
                                  - transformer: {implied_transformer_dropout}""")
        # check upper bound dropout:
        if cfg.mc_dropout and max_implied_dropout >= 1.0:
            raise RuntimeError(f"""The maximal dropout-rate is 1. Please check your dropout-settings:
                               The current implied dropout-rates are:
                                  - model: {cfg.output_dropout}
                                  - statics_embedding: {implied_statics_embedding}
                                  - dynamics_embedding: {implied_dynamics_embedding}
                                  - statics_forecast_embedding: {implied_forecast_statics_embedding}
                                  - dynamics_forecast_embedding: {implied_forecast_dynamics_embedding}
                                  - statics_hindcast_embedding: {implied_hindcast_statics_embedding}
                                  - dynamics_hindcast_embedding: {implied_hindcast_dynamics_embedding}
                                  - transformer: {implied_transformer_dropout}""")

        # assign setup properties:
        self.cfg = cfg
        self.device = next(model.parameters()).device
        self.number_of_targets = len(cfg.target_variables)
        self.mc_dropout = cfg.mc_dropout
        self.predict_last_n = cfg.predict_last_n

        # determine appropriate frequency suffix:
        if self.cfg.use_frequencies:
            self.freq_suffixes = [f'_{freq}' for freq in cfg.use_frequencies]
        else:
            self.freq_suffixes = ['']

        self.batch_size_data = data[f'y{self.freq_suffixes[0]}'].shape[0]

    def _get_frequency_last_n(self, freq_suffix: str):
        if isinstance(self.predict_last_n, int):
            frequency_last_n = self.predict_last_n
        else:
            frequency_last_n = self.predict_last_n[freq_suffix[1:]]
        return frequency_last_n


def sample_mcd(model: 'BaseModel', data: Dict[str, torch.Tensor], n_samples: int,
               scaler: Dict[str, Union[pd.Series, xarray.Dataset]]) -> Dict[str, torch.Tensor]:
    """MC-Dropout based point predictions sampling.

    Naive sampling. This function does `n_samples` forward passes for each sample in the batch. Currently it is 
    only useful for models with dropout, to perform MC-Dropout sampling. 
    Note: Calling this function will force the model to train mode (`model.train()`) and not set it back to its original
    state. 

    The negative sample handling currently supports (a) 'clip' for directly clipping sample_points at zero and (b) 
    'truncate' for resampling sample_points that are below zero. The mode can be defined by the config argument 
    'negative_sample_handling'.

    Parameters
    ----------
    model : BaseModel
        A model with a non-probabilistic head.
    data : Dict[str, torch.Tensor]
        Dictionary, containing input features as key-value pairs.
    n_samples : int
        Number of samples to generate for each input sample.
    scaler : Dict[str, Union[pd.Series, xarray.Dataset]]
        Scaler of the run.

    Returns
    -------
    Dict[str, torch.Tensor]
        Dictionary, containing the sampled model outputs for the `predict_last_n` (config argument) time steps of 
        each frequency.
    """
    setup = _SamplingSetup(model, data, model.cfg.head)

    # force model into train mode for mc_dropout:
    if setup.mc_dropout:
        model.train()

    # sample for different frequencies and targets:
    samples = {}
    for freq_suffix in setup.freq_suffixes:
        sample_points = []
        frequency_last_n = setup._get_frequency_last_n(freq_suffix=freq_suffix)

        for nth_target in range(setup.number_of_targets):
            # unbound sampling:
            def _sample_values(ids: List[int]) -> torch.Tensor:
                # The ids are used for location-specific resampling for 'truncation' in '_handle_negative_values'
                target_values = torch.zeros(len(ids), frequency_last_n, n_samples)
                for i in range(n_samples):  # forward-pass for each frequency separately to guarantee independence
                    prediction = model(data)
                    value_buffer = prediction[f'y_hat{freq_suffix}'][:, -frequency_last_n:, 0]
                    target_values[ids, -frequency_last_n:, i] = value_buffer.detach().cpu()
                return target_values

            ids = list(range(data[f'x_d{freq_suffix}'].shape[0]))
            values = _sample_values(ids)

            # bind values and add to sample_points:
            values = _handle_negative_values(setup.cfg, values, _sample_values, scaler, nth_target)
            sample_points.append(values)

        # add sample_points to dictionary of samples:
        freq_key = f'y_hat{freq_suffix}'
        samples.update({freq_key: torch.stack(sample_points, 2)})

    return samples


def sample_gmm(model: 'BaseModel', data: Dict[str, torch.Tensor], n_samples: int,
               scaler: Dict[str, Union[pd.Series, xarray.Dataset]]) -> Dict[str, torch.Tensor]:
    """Sample point predictions with the Gaussian Mixture (GMM) head.

    This function generates `n_samples` GMM sample points for each entry in the batch. Concretely, the model is 
    executed once (forward pass) and then the sample points are generated by sampling from the resulting mixtures. 
    Good references for learning about GMMs are [#]_ and [#]_. 

    The negative sample handling currently supports (a) 'clip' for directly clipping sample_points at zero and 
     (b) 'truncate' for resampling sample_points that are below zero. The mode can be defined by the config argument 
     'negative_sample_handling'.
     
    Note: If the config setting 'mc_dropout' is true this function will force the model to train mode (`model.train()`) 
    and not set it back to its original state. 

    Parameters
    ----------
    model : BaseModel
        A model with a GMM head.
    data : Dict[str, torch.Tensor]
        Dictionary, containing input features as key-value pairs.
    n_samples : int
        Number of samples to generate for each input sample.
    scaler : Dict[str, Union[pd.Series, xarray.Dataset]]
        Scaler of the run.

    Returns
    -------
    Dict[str, torch.Tensor]
        Dictionary, containing the sampled model outputs for the `predict_last_n` (config argument) time steps of 
        each frequency. 

    References
    ----------
    .. [#] C. M. Bishop: Mixture density networks. 1994.
    .. [#] D. Ha: Mixture density networks with tensorflow. blog.otoro.net, 
           URL: http://blog.otoro.net/2015/11/24/mixture-density-networks-with-tensorflow, 2015.
    """
    setup = _SamplingSetup(model, data, "gmm")

    # force model into train mode if mc_dropout:
    if setup.mc_dropout:
        model.train()

    # make predictions:
    pred = model(data)

    # sample for different frequencies:
    samples = {}
    for freq_suffix in setup.freq_suffixes:
        # get predict_last_n for the given the mode:
        frequency_last_n = setup._get_frequency_last_n(freq_suffix=freq_suffix)

        # initialize sample_points tensor for sampling:
        sample_points = torch.zeros((setup.batch_size_data, frequency_last_n, setup.number_of_targets, n_samples))
        sample_points *= torch.tensor(float('nan'))  # set initial sample_points to nan

        # GMM has 3 parts: means (m/mu), variances (s/sigma), and weights (p/pi):
        m, s, p = pred[f'mu{freq_suffix}'], \
                  pred[f'sigma{freq_suffix}'], \
                  pred[f'pi{freq_suffix}']

        for nth_target in range(setup.number_of_targets):
            m_target = _subset_target(m[:, -frequency_last_n:, :], nth_target, setup.cfg.n_distributions)
            s_target = _subset_target(s[:, -frequency_last_n:, :], nth_target, setup.cfg.n_distributions)
            p_target = _subset_target(p[:, -frequency_last_n:, :], nth_target, setup.cfg.n_distributions)

            mask_nan = ~torch.isnan(m_target[:, -1, 0])
            if any(mask_nan):  # skip if the complete mini-batch is invalid
                m_sub = torch.repeat_interleave(m_target[mask_nan, :, :], n_samples, dim=0)
                s_sub = torch.repeat_interleave(s_target[mask_nan, :, :], n_samples, dim=0)
                p_sub = torch.repeat_interleave(p_target[mask_nan, :, :], n_samples, dim=0)

                # sample values, handle negatives and add to sample points:
                values = _sample_gaussian_mixtures(torch.ones(s_sub.shape, dtype=bool), m_sub, s_sub, p_sub)
                values = _handle_negative_values(
                    setup.cfg,
                    values,
                    sample_values=lambda ids: _sample_gaussian_mixtures(ids, m_sub, s_sub, p_sub),
                    scaler=scaler,
                    nth_target=nth_target)
                values = values.view(-1, n_samples, frequency_last_n).permute(0, 2, 1)

                sample_points[mask_nan, :, nth_target, :] = values.detach().cpu()

        # add sample_points to dictionary of samples:
        freq_key = f'y_hat{freq_suffix}'
        samples.update({freq_key: sample_points})
    return samples


def sample_cmal(model: 'BaseModel', data: Dict[str, torch.Tensor], n_samples: int,
                scaler: Dict[str, Union[pd.Series, xarray.Dataset]]) -> Dict[str, torch.Tensor]:
    """Sample point predictions with the Countable Mixture of Asymmetric Laplacians (CMAL) head.

    This function generates `n_samples` CMAL sample points for each entry in the batch. Concretely, the model is 
    executed once (forward pass) and then the sample points are generated by sampling from the resulting mixtures. 
    General information about CMAL can be found in [#]_.

    The negative sample handling currently supports (a) 'clip' for directly clipping sample_points at zero and (b) 
    'truncate' for resampling sample_points that are below zero. The mode can be defined by the config argument 
    'negative_sample_handling'.

    Note: If the config setting 'mc_dropout' is true this function will force the model to train mode (`model.train()`) 
    and not set it back to its original state. 

    Parameters
    ----------
    model : BaseModel
        A model with a CMAL head.
    data : Dict[str, torch.Tensor]
        Dictionary, containing input features as key-value pairs.
    n_samples : int
        Number of samples to generate for each input sample.
    scaler : Dict[str, Union[pd.Series, xarray.Dataset]]
        Scaler of the run.

    Returns
    -------
    Dict[str, torch.Tensor]
        Dictionary, containing the sampled model outputs for the `predict_last_n` (config argument) time steps of 
        each frequency. The shape of the output tensor for each frequency is 
        ``[batch size, predict_last_n, n_samples]``.

    References
    ----------
    .. [#] D.Klotz, F. Kratzert, M. Gauch, A. K. Sampson, G. Klambauer, S. Hochreiter, and G. Nearing: 
        Uncertainty Estimation with Deep Learning for Rainfall-Runoff Modelling. arXiv preprint arXiv:2012.14295,
        2020.
    """
    setup = _SamplingSetup(model, data, "cmal")

    # force model into train mode if mc_dropout
    if setup.mc_dropout:
        model.train()

    # make predictions:
    pred = model(data)

    # sample for different frequencies:
    samples = {}
    for freq_suffix in setup.freq_suffixes:
        # get predict_last_n for the given the mode:
        frequency_last_n = setup._get_frequency_last_n(freq_suffix=freq_suffix)

        # CMAL has 4 parts: means (m/mu), scales (b), asymmetries (t/) and weights (p/pi):
        m = pred[f'mu{freq_suffix}']
        b = pred[f'b{freq_suffix}']
        t = pred[f'tau{freq_suffix}']
        p = pred[f'pi{freq_suffix}']

        sample_points = []
        for nth_target in range(setup.number_of_targets):
            # sampling presets:
            m_target = _subset_target(m[:, -frequency_last_n:, :], nth_target, setup.cfg.n_distributions)
            b_target = _subset_target(b[:, -frequency_last_n:, :], nth_target, setup.cfg.n_distributions)
            t_target = _subset_target(t[:, -frequency_last_n:, :], nth_target, setup.cfg.n_distributions)
            p_target = _subset_target(p[:, -frequency_last_n:, :], nth_target, setup.cfg.n_distributions)

            m_target = torch.repeat_interleave(m_target, n_samples, dim=0)
            b_target = torch.repeat_interleave(b_target, n_samples, dim=0)
            t_target = torch.repeat_interleave(t_target, n_samples, dim=0)
            p_target = torch.repeat_interleave(p_target, n_samples, dim=0)

            # sampling procedure:
            values = torch.zeros((setup.batch_size_data * n_samples, frequency_last_n)).to(setup.device)
            values *= torch.tensor(float('nan'))  # set target sample_points to nan
            for nth_timestep in range(frequency_last_n):

                mask_nan = ~torch.isnan(p_target[:, nth_timestep, 0])
                if any(mask_nan):  # skip if the complete mini-batch is invalid
                    sub_choices = torch.multinomial(p_target[mask_nan, nth_timestep, :], num_samples=1)
                    t_sub = t_target[mask_nan, nth_timestep, :].gather(1, sub_choices)
                    m_sub = m_target[mask_nan, nth_timestep, :].gather(1, sub_choices)
                    b_sub = b_target[mask_nan, nth_timestep, :].gather(1, sub_choices)

                    ids = torch.ones(b_sub.shape, dtype=bool)
                    values_unbound = _sample_asymmetric_laplacians(ids, m_sub, b_sub, t_sub)
                    values[mask_nan, nth_timestep] = _handle_negative_values(
                        setup.cfg,
                        values_unbound,
                        sample_values=lambda ids: _sample_asymmetric_laplacians(ids, m_sub, b_sub, t_sub),
                        scaler=scaler,
                        nth_target=nth_target)

            # add the values to the sample_points:
            values = values.permute(1, 0).reshape(frequency_last_n, -1, n_samples).permute(1, 0, 2)
            values = values.detach().cpu()
            sample_points.append(values)

        # add sample_points to dictionary of samples:
        freq_key = f'y_hat{freq_suffix}'
        samples.update({freq_key: torch.stack(sample_points, 2)})
    return samples


def sample_umal(model: 'BaseModel', data: Dict[str, torch.Tensor], n_samples: int,
                scaler: Dict[str, Union[pd.Series, xarray.Dataset]]) -> Dict[str, torch.Tensor]:
    """Sample point predictions with the Uncountable Mixture of Asymmetric Laplacians (UMAL) head.

    This function generates `n_samples` UMAL sample points for each entry in the batch. Concretely, the model is 
    executed once (forward pass) and then the sample points are generated by sampling from the resulting mixtures. 
    Details about the UMAL approach can be found in [#]_.

    The negative sample handling currently supports (a) 'clip' for directly clipping sample_points at zero and (b) 
    'truncate' for resampling sample_points that are below zero. The mode can be defined by the config argument 
    'negative_sample_handling'.
    
    Note: If the config setting 'mc_dropout' is true this function will force the model to train mode (`model.train()`) 
    and not set it back to its original state. 

    Parameters
    ----------
    model : BaseModel
        A model with an UMAL head.
    data : Dict[str, torch.Tensor]
        Dictionary, containing input features as key-value pairs.
    n_samples : int
        Number of samples to generate for each input sample.
    scaler : Dict[str, Union[pd.Series, xarray.Dataset]]
        Scaler of the run.

    Returns
    -------
    Dict[str, torch.Tensor]
        Dictionary containing the sampled model outputs for the `predict_last_n` (config argument) time steps of 
        each frequency.

    References
    ----------
    .. [#] A. Brando, J. A. Rodriguez, J. Vitria, and A. R. Munoz: Modelling heterogeneous distributions 
        with an Uncountable Mixture of Asymmetric Laplacians. Advances in Neural Information Processing Systems, 
        pp. 8838-8848, 2019.
    """
    setup = _SamplingSetup(model, data, "umal")

    # force model into train mode if mc_dropout:
    if setup.mc_dropout:
        model.train()

    # n_taus expands the batch by itself and adds a sampled tau as input (new_batch_size = n_taus*batch_size):
    if 'y_extended' not in data:
        data = model.pre_model_hook(data, is_train=False)

    # make predictions:
    pred = model(data)

    # sample:
    samples = {}
    for freq_suffix in setup.freq_suffixes:
        # get predict_last_n for the given the mode:
        frequency_last_n = setup._get_frequency_last_n(freq_suffix=freq_suffix)

        # UMAL has 2 parts: means (m/mu), scales (b); the tau is randomly chosen:
        m = pred[f'mu{freq_suffix}']
        b = pred[f'b{freq_suffix}']
        t = data[f'tau{freq_suffix}']

        # sampling presets:
        m_wide = torch.cat(m[:, -frequency_last_n:, :].split(setup.batch_size_data, 0), 2)
        b_wide = torch.cat(b[:, -frequency_last_n:, :].split(setup.batch_size_data, 0), 2)

        # for now we just use a single tau for all targets:
        t_target = torch.cat(t[:, -frequency_last_n:, :].split(setup.batch_size_data, 0), 2)

        # sample over targets:
        sample_points = torch.zeros((setup.batch_size_data, frequency_last_n, setup.number_of_targets, n_samples))
        sample_points *= torch.tensor(float('nan'))  # set initial sample_points to nan
        for nth_target in range(setup.number_of_targets):
            # sampling presets:
            m_target = _subset_target(m_wide[:, -frequency_last_n:, :], nth_target, setup.cfg.n_taus)
            b_target = _subset_target(b_wide[:, -frequency_last_n:, :], nth_target, setup.cfg.n_taus)

            # sample over n_samples:
            for nth_sample in range(n_samples):
                sub_choice = np.random.randint(0, setup.cfg.n_taus)

                mask_nan = ~torch.isnan(m_target[:, 0, 0])
                if any(mask_nan):  # skip computation if entire mini-batch is invalid
                    m_sub = m_target[mask_nan, :, sub_choice]
                    b_sub = b_target[mask_nan, :, sub_choice]
                    t_sub = t_target[mask_nan, :, sub_choice]

                    ids = torch.ones(b_sub.shape, dtype=bool)
                    values_unbound = _sample_asymmetric_laplacians(ids, m_sub, b_sub, t_sub)
                    values = _handle_negative_values(
                        setup.cfg,
                        values_unbound,
                        sample_values=lambda ids: _sample_asymmetric_laplacians(ids, m_sub, b_sub, t_sub),
                        scaler=scaler,
                        nth_target=nth_target)

                    # add values to sample_points:
                    values = values.detach().cpu().unsqueeze(1)
                    sample_points[mask_nan, -frequency_last_n:, nth_target, nth_sample] = values

        # add sample_points to dictionary of samples:
        freq_key = f'y_hat{freq_suffix}'
        samples.update({freq_key: sample_points})
    return samples

def umal_extend_batch(data: Dict[str, torch.Tensor], cfg: Config, n_taus: int = 1, extend_y: bool = False) \
        -> Dict[str, torch.Tensor]:
    """This function extends the batch for the usage in UMAL (see: [#]_). 
    
    UMAL makes an MC approximation to a mixture integral by sampling random asymmetry parameters (tau). This can be 
    parallelized by expanding the batch for each tau.  
    
    Parameters
    ----------
    data : Dict[str, torch.Tensor]
        Dictionary, containing input features as key-value pairs.
    cfg : Config
        The run configuration.
    n_taus : int
        Number of taus to expand the batch.
    extend_y : bool
        Option to also extend the labels/y. 
    
    Returns
    -------
    Dict[str, torch.Tensor]
        Dictionary, containing expanded input features and tau samples as key-value pairs.

    References
    ----------
    .. [#] A. Brando, J. A. Rodriguez, J. Vitria, and A. R. Munoz: Modelling heterogeneous distributions 
        with an Uncountable Mixture of Asymmetric Laplacians. Advances in Neural Information Processing Systems, 
        pp. 8838-8848, 2019.
    """
    # setup:
    if cfg.use_frequencies:
        freq_suffixes = [f'_{freq}' for freq in cfg.use_frequencies]
    else:
        freq_suffixes = ['']

    for freq_suffix in freq_suffixes:
        batch_size, seq_length, input_size = data[f'x_d{freq_suffix}'].shape

        if isinstance(cfg.predict_last_n, int):
            predict_last_n = cfg.predict_last_n
        else:
            predict_last_n = cfg.predict_last_n[freq_suffix[1:]]

        # sample tau within [tau_down, tau_up] and add to data:
        tau = (cfg.tau_up - cfg.tau_down) * torch.rand(batch_size * n_taus, 1, 1) + cfg.tau_down
        tau = tau.repeat(1, seq_length, 1)  # in our convention tau remains the same over all inputs
        tau = tau.to(data[f'x_d{freq_suffix}'].device)
        data[f'tau{freq_suffix}'] = tau[:, -predict_last_n:, :]

        # extend dynamic inputs with tau and expand batch:
        x_d = data[f'x_d{freq_suffix}'].repeat(n_taus, 1, 1)
        data[f'x_d{freq_suffix}'] = torch.cat([x_d, tau], dim=-1)
        data[f'y_extended{freq_suffix}'] = data[f'y{freq_suffix}'].repeat(n_taus, 1, 1)
        if f'x_s{freq_suffix}' in data:
            data[f'x_s{freq_suffix}'] = data[f'x_s{freq_suffix}'].repeat(n_taus, 1)
        if f'x_one_hot{freq_suffix}' in data:
            data[f'x_one_hot{freq_suffix}'] = data[f'x_one_hot{freq_suffix}'].repeat(n_taus, 1)

    return data


@njit
def bernoulli_subseries_sampler(
    data: np.ndarray,
    missing_fraction: float,
    mean_missing_length: float,
    start_sampling_on: bool = True,
) -> np.ndarray:
    """Samples a timeseries according to a pair of Bernoulli processes.

    The objective is to sample subsequences of a given timeseries under two criteria:
        1)  Expected long-term sample ratio (i.e., the total fraction of points in a time series 
            that are not sampled): `missing_fraction`.
        2)  Expected length of continuous subsequences sampled from the timeseries:
            `mean_missing_length`.
    
    This is done by sampling two Bernoulli processes with different rate parameters. One
    process samples on-shifts and one process samples off-shifts. An 'on-shift' occurs
    when the state of the sampler transitions from 'off' (not sampling) to 'on' (sampling),
    and vice-versa. The rate parameters for the on-shift and off-shift processes are
    derived from the input parameters explained above.
 
    Parameters
    ----------
    data : np.ndarray
        Time series data to be sampled. Must be (N, 1) where N is the length of the timeseries.
    missing_fraction : float
        Expected total fraction of points in a time series that are not sampled.
    mean_missing_length : float
        Expected length of continuous subsequences of un-sampled data from the timeseries.
    start_sampling_on: bool
        Whether to start with the sampler turned "on" (True) or "off" (False) at the first
        timestep of the timeseries.

    Returns
    -------
    A copy of the timeseries with NaN's replacing elements that were not sampled.
    """
    # Check if sampling ratio is one or zero. Avoids a divide-by-zero error.
    if missing_fraction == 0:
        return data
    if missing_fraction == 1:
        return np.full(data.shape, np.nan)

    # Check that the input data is a 1-d timeseries.
    if not (data.ndim == 1 or (data.ndim == 2 and data.shape[-1] == 1)):
        raise ValueError('Shape of timeseries data must be N or (N, 1).')

    # Check that the distribution parameters make sense.
    if mean_missing_length < missing_fraction / (1 - missing_fraction):
        raise ValueError('Incompatible distribution parameters in timeseries sampling. Must be: ',
                         'mean_missing_length >= missing_fraction / (1-missing_fraction).')
    if missing_fraction < 0 or missing_fraction > 1:
        raise ValueError('Missing fraction must be in [0,1]')

    if mean_missing_length <= 0:
        raise ValueError('Mean missing length must be > 0.')

    # Derive Bernoulli rate parameters.
    on_shift_rate = 1 / mean_missing_length
    off_shift_rate = on_shift_rate * missing_fraction / (1 - missing_fraction)

    # Initialize storage for the samples.
    sampled_data = np.full(data.shape, np.nan)
    sampled_data[0] = data[0]
    if not start_sampling_on:
        sampled_data[0] = np.nan

    # Bernoulli sampling.
    up_switches = np.random.binomial(n=1, p=on_shift_rate, size=data.shape)
    down_switches = np.random.binomial(n=1, p=off_shift_rate, size=data.shape)

    # Sampling -> stochastic process.
    for n in range(1, len(data)):
        if np.isnan(sampled_data[n - 1]):
            if up_switches[n]:
                sampled_data[n] = data[n]
        else:
            if not down_switches[n]:
                sampled_data[n] = data[n]

    return sampled_data



import math
from collections import defaultdict

import re
from typing import Dict, List

import torch
import torch.nn as nn

from neuralhydrology.modelzoo.basemodel import BaseModel
from neuralhydrology.modelzoo.inputlayer import InputLayer
from neuralhydrology.modelzoo.head import get_head
from neuralhydrology.utils.config import Config


class ARLSTM(BaseModel):
    """An autoregressive LSTM.

    This model assumes that the *last* entry in dynamic inpus (x_d) is an observation that is supposed to match
    target data lagged by an integer number of timesteps. If this data is missing (NaN), it is substituted 
    for the model prediction at the same lag. The model adds an extra dynamic input that serves as a binary 
    flag to indicate whether the autoregressive input at a particular timestep is from observation vs. simulation.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    """

    def __init__(self, cfg: Config):
        super(ARLSTM, self).__init__(cfg=cfg)

        self._ar_shift = self._get_ar_shift()
        self._num_ar_inputs = len(self.cfg.autoregressive_inputs)
        if self.output_size != self._num_ar_inputs:
            raise ValueError('The AR-LSTM currently only works if all outputs are used as lagged inputs.')

        self.embedding_net = InputLayer(cfg)

        # increase input size for binary flag
        input_size = self.embedding_net.output_size + self._num_ar_inputs
        self.cell = nn.LSTM(input_size=input_size, hidden_size=cfg.hidden_size)

        self.dropout = nn.Dropout(p=cfg.output_dropout)

        self.head = get_head(cfg=cfg, n_in=cfg.hidden_size, n_out=self.output_size)

        self._reset_parameters()

    def _get_ar_shift(self) -> int:
        shifts = set()
        for input in self.cfg.autoregressive_inputs:
            capture = re.compile(r'^(.*)_shift(\d+)$').search(input)
            if not capture:
                raise ValueError('Autoregressive inputs must be a shifted variable with form <variable>_shift<lag> ',
                                f'where <lag> is an integer. Instead got: {input}.')
            if not capture[1] in self.cfg.target_variables:
                raise ValueError('Autoregressive inputs must be a shifted target variable. ',
                                f'Instead got a shifted version of: {capture[1]}.')
            shifts.add(int(capture[2]))
        if len(shifts) > 1:
            raise ValueError('Only one AR shift is allowed currently. All autoregressive inputs must use the same shift.')
        shift = shifts.pop()
        if shift <= 0:
            raise ValueError('Autoregressive inputs must be shifted by at least one timestep.')
        return shift    
    
    def _reset_parameters(self):
        """Special initialization of certain model weights."""
        if self.cfg.initial_forget_bias is not None:
            self.cell.bias_hh_l0.data[self.cfg.hidden_size:2 * self.cfg.hidden_size] = self.cfg.initial_forget_bias

    def forward(self,
                data: Dict[str, torch.Tensor],
                h_0: torch.Tensor = None,
                c_0: torch.Tensor = None) -> Dict[str, torch.Tensor]:
        """Perform a forward pass on the Autoregressive LSTM model.

        Parameters
        ----------
        data : Dict[str, torch.Tensor]
            Dictionary, containing input features as key-value pairs.

        Returns
        -------
        Dict[str, torch.Tensor]
            Model outputs and intermediate states as a dictionary.
                - `y_hat`: model predictions of shape [batch size, sequence length, number of target variables].
                - `lstm_output`: full timeseries of the hidden states from the lstm [batch size, sequence length, hidden size].
                - `h_n`: hidden state at the last time step of the sequence of shape [batch size, 1, hidden size].
                - `c_n`: cell state at the last time step of the sequence of shape [batch size, 1, hidden size].
        """
        # possibly pass dynamic and static inputs through embedding layers, then concatenate them
        x_d = self.embedding_net(data)

        _, batch_size, _ = x_d.size()

        if h_0 is None:
            h_0 = x_d.new_zeros((1, batch_size, self.cfg.hidden_size))
        if c_0 is None:
            c_0 = x_d.new_zeros((1, batch_size, self.cfg.hidden_size))

        # initialize flags for autoregressive inputs -- all flags start at 0, indicating obs
        # flags are appended as the last input
        ar_flags = x_d.new_zeros((batch_size, self._num_ar_inputs))
        
        # initialize the 'last' prediction to substitute for any missing AR values in the first timesteps
        last_prediction = x_d.new_zeros((self._ar_shift, batch_size, self.output_size))

        # initialize dictionary to store the output
        lstm_output = []
        y_hat = []

        # manually loop through timesteps
        for x_t in x_d:

            # find locations of missing data (NaN's) and replace with last predictions
            x_embd = x_t[:, :-self._num_ar_inputs]
            x_ar = x_t[:, -self._num_ar_inputs:].clone()
            replace_indexes = torch.isnan(x_ar)
            x_ar[replace_indexes] = last_prediction[-1, replace_indexes]
            ar_flags[:, :] = 0
            ar_flags[replace_indexes] = 1

            # one timestep of lstm
            cell_inputs = torch.unsqueeze(torch.concat([x_embd, x_ar, ar_flags], -1), 0)
            cell_output, (h_0, c_0) = self.cell(cell_inputs, (h_0, c_0))

            # append all timestep output to the output dictionary
            lstm_output.append(cell_output.transpose(0, 1))

            # store the last prediction
            last_prediction[1:] = last_prediction[:-1,].clone()
            prediction = torch.squeeze(self.head(self.dropout(h_0.transpose(0, 1)))['y_hat'], dim=1)
            last_prediction[0] = prediction
            y_hat.append(prediction)

        # stack all ouptuts to sizes in function doc
        pred = {
            'lstm_output': torch.concat(lstm_output, 1), 
            'h_n': h_0.transpose(0, 1), 
            'c_n': c_0.transpose(0, 1),
            'y_hat': torch.stack(y_hat, 1),
        }
        return pred



from typing import Dict

import torch
import torch.nn as nn

from neuralhydrology.utils.config import Config
from neuralhydrology.modelzoo.basemodel import BaseModel
from neuralhydrology.modelzoo.fc import FC
from neuralhydrology.modelzoo.head import get_head
from neuralhydrology.modelzoo.inputlayer import InputLayer

class HandoffForecastLSTM(BaseModel):
    """An encoder/decoder LSTM model class used for forecasting.

    This is a forecasting model that uses a state-handoff to transition from a hindcast sequence model to a forecast 
    sequence (LSTM) model. The hindcast model is run from the past up to present (the issue time of the forecast) 
    and then passes the cell state and hidden state of the LSTM into a (nonlinear) handoff network, which is then 
    used to initialize the cell state and hidden state of a new LSTM that rolls out over the forecast period. 
    The handoff network is implemented as a custom FC layer, which can have multiple layers. The handoff network is 
    implemented using the ``state_handoff_network`` config parameter. The hindcast and forecast LSTMs have different 
    weights and biases, different heads, and different embedding networks. The hidden size of the hindcast 
    LSTM is set using the ``hindcast_hidden_size`` config parameter and the hidden size of the forecast LSTM is set 
    using the ``forecast_hidden_size`` config parameter.

    The handoff forecast LSTM model can implement a delayed handoff as well, such that the handoff between the hindcast
    and forecast LSTM occurs prior to the forecast issue time. This is controlled by the ``forecast_overlap`` parameter
    in the config file, and the forecast and hindcast LSTMs will run concurrently for the number of timesteps indicated
    by ``forecast_overlap``. We recommend using the ``ForecastOverlapMSERegularization`` regularization option to
    regularize the loss function by (dis)agreement between the overlapping portion of the hindcast and forecast LSTMs.
    This regularization term can be requested by setting the ``regularization`` parameter in the config file to include
    ``forecast_overlap``.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    
    Raises
    ------
    ValueError if a state_handoff_network is not specified.
    """
    # specify submodules of the model that can later be used for finetuning. Names must match class attributes
    module_parts = ['hindcast_embedding_net', 'forecast_embedding_net', 'hindcast_lstm', 'forecast_lstm', 'hindcast_head', 'forecast_head', 'handoff_net']

    def __init__(self, cfg: Config):
        super(HandoffForecastLSTM, self).__init__(cfg=cfg)

        self.initial_hindcast_seq_length = self.cfg.seq_length - self.cfg.forecast_seq_length

        self.forecast_embedding_net = InputLayer(cfg, embedding_type='forecast')
        self.hindcast_embedding_net = InputLayer(cfg, embedding_type='hindcast')

        self.hindcast_hidden_size = cfg.hindcast_hidden_size
        self.hindcast_lstm = nn.LSTM(
            input_size=self.hindcast_embedding_net.output_size,
            hidden_size=self.hindcast_hidden_size
        )
        self.forecast_hidden_size = cfg.forecast_hidden_size
        self.forecast_lstm = nn.LSTM(
            input_size=self.forecast_embedding_net.output_size,
            hidden_size=self.forecast_hidden_size
        )

        if not cfg.state_handoff_network:
            raise ValueError('The handoff forecast LSTM requires a state handoff network specified in the config file.')

        self.handoff_net = FC(
            input_size=self.hindcast_hidden_size*2,
            hidden_sizes=cfg.state_handoff_network['hiddens'],
            activation=cfg.state_handoff_network['activation'],
            dropout=cfg.state_handoff_network['dropout']
        )
        self.handoff_linear =  FC(
            input_size=cfg.state_handoff_network['hiddens'][-1],
            hidden_sizes=[self.forecast_hidden_size*2],
            activation='linear',
            dropout=0.0
        )

        self.dropout = nn.Dropout(p=cfg.output_dropout)

        self.hindcast_head = get_head(cfg=cfg, n_in=self.hindcast_hidden_size, n_out=self.output_size)
        self.forecast_head = get_head(cfg=cfg, n_in=self.forecast_hidden_size, n_out=self.output_size)

        self._reset_parameters()

    def _reset_parameters(self):
        """Special initialization of certain model weights."""
        if self.cfg.initial_forget_bias is not None:
            self.hindcast_lstm.bias_hh_l0.data[self.hindcast_hidden_size:2 * self.hindcast_hidden_size] = self.cfg.initial_forget_bias
            self.forecast_lstm.bias_hh_l0.data[self.forecast_hidden_size:2 * self.forecast_hidden_size] = self.cfg.initial_forget_bias

    def forward(self, data: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform a forward pass on the HandoffForecastLSTM model.

        Parameters
        ----------
        data : Dict[str, torch.Tensor]
            Dictionary, containing input features as key-value pairs.

        Returns
        -------
        Dict[str, torch.Tensor]
            Model outputs and intermediate states as a dictionary.
                - lstm_output_hindcast: Output sequence from the hindcast LSTM.
                - lstm_output_hindcast_overlap: Output sequence from the hindcast model over the overlap period between forecast and hindcast LSTMs.
                - lstm_output_forecast_overlap: Output sequence from the forecast model over the overlap period between forecast and hindcast LSTMs.
                - lstm_output_forecast: Output sequence from the forecast LSTM.
                - y_forecast: Predictions (after head layer) over the forecast period.
                - y_forecast_overlap: Predictions from the forecast model over the overlap period between forecast and hindcast LSTMs.
                - y_hindcast_overlap: Predictions from the hindcast model over the overlap period between forecast and hindcast LSTMs.
                - y_hindcast: Predictions over the hindcast period.
                - h_n_hindcast: Final hidden state of the hindcast model.
                - c_n_hindcast: Final cell state of the hindcast model.
                - h_n_handoff: Initial hidden state of the forecast model.
                - c_n_handoff: Initial cell state of the forecast model.
                - h_n_forecast: Finall hidden state of the forecast model.
                - c_n_forecast: Final cell state of the forecast model.
                - y_hat: Predictions over the sequence from the head layer. This is a concatenation of hindcast and forecast, and takes from hindcast for the overlap portion.
        """
        # possibly pass dynamic and static inputs through embedding layers, then concatenate them
        x_h = self.hindcast_embedding_net(data)
        x_f = self.forecast_embedding_net(data)

        # run the hindcast lstm
        lstm_output_hindcast, (h_n_hindcast, c_n_hindcast) = self.hindcast_lstm(x_h[:self.initial_hindcast_seq_length, ...])
        lstm_output_hindcast = lstm_output_hindcast.transpose(0, 1)
        if x_h.shape[0] > self.initial_hindcast_seq_length:
            lstm_output_hindcast_overlap, _ = self.hindcast_lstm(x_h[self.initial_hindcast_seq_length:, ...], (h_n_hindcast, c_n_hindcast))
            lstm_output_hindcast_overlap = lstm_output_hindcast_overlap.transpose(0, 1)
        else:
            lstm_output_hindcast_overlap = None

        # handoff initial state to forecast lstm
        x = self.handoff_net(torch.cat([h_n_hindcast, c_n_hindcast], -1))
        initial_state = self.handoff_linear(x)
        h_n_handoff, c_n_handoff = initial_state.chunk(2, -1)
        h_n_handoff = h_n_handoff.contiguous()
        c_n_handoff = c_n_handoff.contiguous()

        # run forecast lstm
        lstm_output_forecast, (h_n_forecast, c_n_forecast) = self.forecast_lstm(x_f, (h_n_handoff, c_n_handoff))
        lstm_output_forecast = lstm_output_forecast.transpose(0, 1)
        lstm_output_forecast_overlap = lstm_output_forecast[:, :self.cfg.forecast_overlap, :]
        lstm_output_forecast = lstm_output_forecast[:, self.cfg.forecast_overlap:, :]

        # run heads for hindcast and forecast
        y_hindcast = self.hindcast_head(self.dropout(lstm_output_hindcast))
        y_forecast = self.forecast_head(self.dropout(lstm_output_forecast))
        if x_h.shape[0] > self.initial_hindcast_seq_length:
            y_hindcast_overlap = self.hindcast_head(self.dropout(lstm_output_hindcast_overlap))
            y_forecast_overlap = self.forecast_head(self.dropout(lstm_output_forecast_overlap))
            pred = {key: torch.cat([y_hindcast[key], y_hindcast_overlap[key], y_forecast[key]], dim=1) for key in y_hindcast}
        else:
            pred = {key: torch.cat([y_hindcast[key], y_forecast[key]], dim=1) for key in y_hindcast}
            y_hindcast_overlap, y_forecast_overlap = None, None

        # reshape to [batch_size, seq, n_hiddens]
        h_n_hindcast = h_n_hindcast.transpose(0, 1)
        c_n_hindcast = c_n_hindcast.transpose(0, 1)
        h_n_handoff = h_n_handoff.transpose(0, 1)
        c_n_handoff = c_n_handoff.transpose(0, 1)
        h_n_forecast = h_n_forecast.transpose(0, 1)
        c_n_forecast = c_n_forecast.transpose(0, 1)

        pred.update(
            {
                'lstm_output_hindcast': lstm_output_hindcast,
                'lstm_output_hindcast_overlap': lstm_output_hindcast_overlap,
                'lstm_output_forecast_overlap': lstm_output_forecast_overlap,
                'lstm_output_forecast': lstm_output_forecast,

                'y_forecast': y_forecast,
                'y_forecast_overlap': y_forecast_overlap,
                'y_hindcast_overlap': y_hindcast_overlap,
                'y_hindcast': y_hindcast,

                'h_n_hindcast': h_n_hindcast,
                'c_n_hindcast': c_n_hindcast,

                'h_n_handoff': h_n_handoff,
                'c_n_handoff': c_n_handoff,

                'h_n_forecast': h_n_forecast,
                'c_n_forecast': c_n_forecast,
            }
        )

        return pred



from neuralhydrology.datasetzoo.basedataset import BaseDataset
from neuralhydrology.datasetzoo.camelsaus import CamelsAUS
from neuralhydrology.datasetzoo.camelsbr import CamelsBR
from neuralhydrology.datasetzoo.camelscl import CamelsCL
from neuralhydrology.datasetzoo.camelsgb import CamelsGB
from neuralhydrology.datasetzoo.camelsus import CamelsUS
from neuralhydrology.datasetzoo.caravan import Caravan
from neuralhydrology.datasetzoo.genericdataset import GenericDataset
from neuralhydrology.datasetzoo.hourlycamelsus import HourlyCamelsUS
from neuralhydrology.datasetzoo.lamah import LamaH
from neuralhydrology.utils.config import Config


def get_dataset(cfg: Config,
                is_train: bool,
                period: str,
                basin: str = None,
                additional_features: list = [],
                id_to_int: dict = {},
                scaler: dict = {}) -> BaseDataset:
    """Get data set instance, depending on the run configuration.

    Currently implemented datasets are 'caravan', 'camels_aus', 'camels_br', 'camels_cl', 'camels_gb', 'camels_us', and 
    'hourly_camels_us', as well as the 'generic' dataset class that can be used for any kind of dataset as long as it is
    in the correct format.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    is_train : bool 
        Defines if the dataset is used for training or evaluating. If True (training), means/stds for each feature
        are computed and stored to the run directory. If one-hot encoding is used, the mapping for the one-hot encoding 
        is created and also stored to disk. If False, a `scaler` input is expected and similarly the `id_to_int` input
        if one-hot encoding is used. 
    period : {'train', 'validation', 'test'}
        Defines the period for which the data will be loaded
    basin : str, optional
        If passed, the data for only this basin will be loaded. Otherwise the basin(s) is(are) read from the appropriate
        basin file, corresponding to the `period`.
    additional_features : List[Dict[str, pd.DataFrame]], optional
        List of dictionaries, mapping from a basin id to a pandas DataFrame. This DataFrame will be added to the data
        loaded from the dataset and all columns are available as 'dynamic_inputs', 'evolving_attributes' and
        'target_variables'
    id_to_int : Dict[str, int], optional
        If the config argument 'use_basin_id_encoding' is True in the config and period is either 'validation' or 
        'test', this input is required. It is a dictionary, mapping from basin id to an integer (the one-hot encoding).
    scaler : Dict[str, Union[pd.Series, xarray.DataArray]], optional
        If period is either 'validation' or 'test', this input is required. It contains the centering and scaling
        for each feature and is stored to the run directory during training (train_data/train_data_scaler.yml).

    Returns
    -------
    BaseDataset
        A new data set instance, depending on the run configuration.

    Raises
    ------
    NotImplementedError
        If no data set class is implemented for the 'dataset' argument in the config.
    """
    if cfg.dataset.lower() == "camels_us":
        Dataset = CamelsUS
    elif cfg.dataset.lower() == "camels_gb":
        Dataset = CamelsGB
    elif cfg.dataset.lower() == "camels_aus":
        Dataset = CamelsAUS
    elif cfg.dataset.lower() == "camels_br":
        Dataset = CamelsBR
    elif cfg.dataset.lower() == "hourly_camels_us":
        Dataset = HourlyCamelsUS
    elif cfg.dataset.lower() == "camels_cl":
        Dataset = CamelsCL
    elif cfg.dataset.lower() == "generic":
        Dataset = GenericDataset
    elif cfg.dataset.lower() in ["lamah_a", "lamah_b", "lamah_c"]:
        Dataset = LamaH
    elif cfg.dataset.lower() == "caravan":
        Dataset = Caravan
    else:
        raise NotImplementedError(f"No dataset class implemented for dataset {cfg.dataset}")

    ds = Dataset(cfg=cfg,
                 is_train=is_train,
                 period=period,
                 basin=basin,
                 additional_features=additional_features,
                 id_to_int=id_to_int,
                 scaler=scaler)
    return ds



from functools import reduce
from pathlib import Path
from typing import Dict, List, Union

import numpy as np
import pandas as pd
import xarray

from neuralhydrology.datasetzoo.basedataset import BaseDataset
from neuralhydrology.utils.config import Config


class GenericDataset(BaseDataset):
    """Data set class for the generic dataset that reads data for any region based on common file layout conventions.

    To use this dataset, the data_dir must contain a folder 'time_series' and (if static attributes are used) a folder
    'attributes'. The folder 'time_series' contains one netcdf file (.nc or .nc4) per basin, named '<basin_id>.nc/nc4'.
    The netcdf file has to have one coordinate called `date`, containing the datetime index. The folder 'attributes' 
    contains one or more comma-separated file (.csv) with static attributes, indexed by basin id. Attributes files can 
    be divided into groups of basins or groups of features (but not both, see `genericdataset.load_attributes` for
    more details).

    Note: Invalid values have to be marked as NaN (e.g. using NumPy's np.nan) in the netCDF files and not something like
    -999 for invalid discharge measurements, which is often found in hydrology datasets. If missing values are not 
    marked as NaN's, the GenericDataset will not be able to identify these values as missing data points.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    is_train : bool
        Defines if the dataset is used for training or evaluating. If True (training), means/stds for each feature
        are computed and stored to the run directory. If one-hot encoding is used, the mapping for the one-hot encoding
        is created and also stored to disk. If False, a `scaler` input is expected and similarly the `id_to_int` input
        if one-hot encoding is used.
    period : {'train', 'validation', 'test'}
        Defines the period for which the data will be loaded
    basin : str, optional
        If passed, the data for only this basin will be loaded. Otherwise the basin(s) are read from the appropriate
        basin file, corresponding to the `period`.
    additional_features : List[Dict[str, pd.DataFrame]], optional
        List of dictionaries, mapping from a basin id to a pandas DataFrame. This DataFrame will be added to the data
        loaded from the dataset and all columns are available as 'dynamic_inputs', 'evolving_attributes' and
        'target_variables'
    id_to_int : Dict[str, int], optional
        If the config argument 'use_basin_id_encoding' is True in the config and period is either 'validation' or
        'test', this input is required. It is a dictionary, mapping from basin id to an integer (the one-hot encoding).
    scaler : Dict[str, Union[pd.Series, xarray.DataArray]], optional
        If period is either 'validation' or 'test', this input is required. It contains the centering and scaling
        for each feature and is stored to the run directory during training (train_data/train_data_scaler.yml).
    """

    def __init__(self,
                 cfg: Config,
                 is_train: bool,
                 period: str,
                 basin: str = None,
                 additional_features: List[Dict[str, pd.DataFrame]] = [],
                 id_to_int: Dict[str, int] = {},
                 scaler: Dict[str, Union[pd.Series, xarray.DataArray]] = {}):
        super(GenericDataset, self).__init__(cfg=cfg,
                                             is_train=is_train,
                                             period=period,
                                             basin=basin,
                                             additional_features=additional_features,
                                             id_to_int=id_to_int,
                                             scaler=scaler)

    def _load_basin_data(self, basin: str) -> pd.DataFrame:
        """Load input and output data. """
        df = load_timeseries(data_dir=self.cfg.data_dir, basin=basin)

        return df

    def _load_attributes(self) -> pd.DataFrame:
        """Load static catchment attributes."""
        return load_attributes(self.cfg.data_dir, basins=self.basins)


def load_attributes(data_dir: Path, basins: List[str] = None) -> pd.DataFrame:
    """Load static attributes.

    Parameters
    ----------
    data_dir : Path
        Path to the data directory. This folder must contain an 'attributes' folder with one or multiple csv files.
    basins : List[str], optional
        If passed, return only attributes for the basins specified in this list. Otherwise, the attributes of
        all basins are returned.

    Returns
    -------
    pandas.DataFrame
        Basin-indexed DataFrame, containing the attributes as columns. If the attributes folder contains multiple
        files, they will be concatenated as follows:
        
        (a) if the intersection of basins is non-empty, the files' attributes are concatenated for the intersection of
            basins. The intersection of attributes must be empty in this case.
        (b) if the intersection of basins is empty but the intersection of attributes is not, the files' basins are
            concatenated for the intersection of attributes.
            
        In all other cases, a ValueError is raised.

    Raises
    ------
    FileNotFoundError
        If the attributes folder is not found or does not contain any csv files.
    ValueError
        If an attributes file contains duplicate basin or attribute names, multiple files are found that have no
        overlap, or there are no attributes for a basin specified in `basins`.
    """
    attributes_path = data_dir / 'attributes'
    if not attributes_path.exists():
        raise FileNotFoundError(f"Attributes folder not found at {attributes_path}")

    files = list(attributes_path.glob('*.csv'))
    if not files:
        raise FileNotFoundError('No attributes files found')

    # Read-in attributes into one big dataframe. Sort by both axes so we can check for identical axes.
    dfs = []
    for f in files:
        df = pd.read_csv(f, dtype={0: str})  # make sure we read the basin id as str
        df = df.set_index(df.columns[0]).sort_index(axis=0).sort_index(axis=1)
        if df.index.has_duplicates or df.columns.has_duplicates:
            raise ValueError(f'Attributes file {f} contains duplicate basin ids or features.')
        dfs.append(df)

    if len(dfs) == 1:
        df = dfs[0]
    else:
        if len(reduce(lambda idx, other_idx: idx.intersection(other_idx), (df.index for df in dfs))) > 0:
            # basin intersection is non-empty -> concatenate attributes, keep intersection of basins
            if np.any(np.unique(np.concatenate([df.columns for df in dfs]), return_counts=True)[1] > 1):
                raise ValueError('If attributes dataframes refer to the same basins, no attribute name may occur '
                                 'multiple times across the different attributes files.')
            concat_axis = 1
        elif len(reduce(lambda cols, other_cols: cols.intersection(other_cols), (df.columns for df in dfs))) > 0:
            # attributes intersection is non-empty -> concatenate basins, keep intersection of attributes
            # no need to check for basin duplicates, since then we'd have had a non-empty basin intersection.
            concat_axis = 0
        else:
            raise ValueError('Attribute files must overlap on either the index or the columns.')

        df = pd.concat(dfs, axis=concat_axis, join='inner')

    if basins:
        if any(b not in df.index for b in basins):
            raise ValueError('Some basins are missing static attributes.')
        df = df.loc[basins]

    return df


def load_timeseries(data_dir: Path, basin: str) -> pd.DataFrame:
    """Load time series data from netCDF files into pandas DataFrame.

    Parameters
    ----------
    data_dir : Path
        Path to the data directory. This folder must contain a folder called 'time_series' containing the time series
        data for each basin as a single time-indexed netCDF file called '<basin_id>.nc/nc4'.
    basin : str
        The basin identifier.

    Returns
    -------
    pd.DataFrame
        Time-indexed DataFrame containing the time series data as stored in the netCDF file.

    Raises
    ------
    FileNotFoundError
        If no netCDF file exists for the specified basin.
    ValueError
        If more than one netCDF file is found for the specified basin.
    """
    files_dir = data_dir / "time_series"
    netcdf_files = list(files_dir.glob("*.nc4"))
    netcdf_files.extend(files_dir.glob("*.nc"))
    netcdf_file = [f for f in netcdf_files if f.stem == basin]
    if len(netcdf_file) == 0:
        raise FileNotFoundError(f"No netCDF file found for basin {basin} in {files_dir}")
    if len(netcdf_file) > 1:
        raise ValueError(f"Multiple netCDF files found for basin {basin} in {files_dir}")

    xr = xarray.open_dataset(netcdf_file[0])
    return xr.to_dataframe()



from typing import Dict
import torch
import torch.nn as nn
from neuralhydrology.utils.config import Config
from neuralhydrology.modelzoo.basemodel import BaseModel
from neuralhydrology.modelzoo.baseconceptualmodel import BaseConceptualModel
from neuralhydrology.modelzoo.inputlayer import InputLayer
from neuralhydrology.modelzoo.shm import SHM


class HybridModel(BaseModel):
    """Wrapper to combine a deep learning model with a conceptual hydrological models. 
    
    In the current implementation, the deep learning model is always an LSTM. The conceptual model is configurable using the config
    argument `conceptual_model`. Currently supported is `['SHM']`.
    
    Parameters
    ----------
    cfg : Config
        The run configuration.
    """

    def __init__(self, cfg: Config):
        super(HybridModel, self).__init__(cfg=cfg)

        self.embedding_net = InputLayer(cfg)

        self.lstm = nn.LSTM(input_size=self.embedding_net.output_size, hidden_size=cfg.hidden_size)

        self.conceptual_model = HybridModel._get_conceptual_model(cfg=cfg)

        self.linear = nn.Linear(in_features=cfg.hidden_size, out_features=len(self.conceptual_model.parameter_ranges))

        self._reset_parameters()

    def _reset_parameters(self):
        """Special initialization of certain model weights."""
        if self.cfg.initial_forget_bias is not None:
            self.lstm.bias_hh_l0.data[self.cfg.hidden_size:2 * self.cfg.hidden_size] = self.cfg.initial_forget_bias

    def forward(self, data: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform a forward pass on the model.

        Parameters
        ----------
        data : Dict[str, torch.Tensor]
            Dictionary, containing input features as key-value pairs.

        Returns
        -------
        Dict[str, torch.Tensor]
            Model outputs, dynamic parameters and intermediate states coming from the conceptual model
        """
       
        # run lstm
        x_d = self.embedding_net(data)
        lstm_output, _ = self.lstm(input=x_d)
        lstm_output = lstm_output.transpose(0, 1)  # reshape to [batch_size, seq, n_hiddens]

        # map lstm outputs to the dimension of the conceptual model´s parameters
        lstm_out = lstm_output[:, self.cfg.warmup_period:, :]
        lstm_out = self.linear(lstm_out)

        # get predictions
        pred = self.conceptual_model(x_conceptual=data['x_d_c'][:, self.cfg.warmup_period:, :], lstm_out=lstm_out)

        return pred

    @staticmethod
    def _get_conceptual_model(cfg: Config) -> BaseConceptualModel:
        """Get conceptual model, depending on the run configuration.

        Parameters
        ----------
        cfg : Config
            The run configuration.

        Returns
        -------
        BaseConceptualModel
            A new conceptual model instance of the type specified in the config.
        """
        if cfg.conceptual_model.lower() == "shm":
            conceptual_model = SHM(cfg=cfg)
        else:
            raise NotImplementedError(f"{cfg.conceptual_model} not implemented or not linked in `_get_conceptual_model()`")

        return conceptual_model



from typing import Dict, Tuple

import torch
import torch.nn as nn

from neuralhydrology.modelzoo.basemodel import BaseModel
from neuralhydrology.modelzoo.inputlayer import InputLayer
from neuralhydrology.modelzoo.head import get_head
from neuralhydrology.utils.config import Config


class EALSTM(BaseModel):
    """Entity-Aware LSTM (EA-LSTM) model class.

    This model has been proposed by Kratzert et al. [#]_ as a variant of the standard LSTM. The main difference is that
    the input gate of the EA-LSTM is modulated using only the static inputs, while the dynamic (time series) inputs are
    used in all other parts of the model (i.e. forget gate, cell update gate and output gate).
    To control the initial forget gate bias, use the config argument `initial_forget_bias`. Often it is useful to set 
    this value to a positive value at the start of the model training, to keep the forget gate closed and to facilitate
    the gradient flow. 
    The `EALSTM` class does only support single timescale predictions. Use `MTSLSTM` to train an LSTM-based model and 
    get predictions on multiple temporal resolutions at the same time.

    Parameters
    ----------
    cfg : Config
        The run configuration.
        
    References
    ----------
    .. [#] Kratzert, F., Klotz, D., Shalev, G., Klambauer, G., Hochreiter, S., and Nearing, G.: Towards learning 
        universal, regional, and local hydrological behaviors via machine learning applied to large-sample datasets, 
        Hydrol. Earth Syst. Sci., 23, 5089-5110, https://doi.org/10.5194/hess-23-5089-2019, 2019.
    """
    # specify submodules of the model that can later be used for finetuning. Names must match class attributes
    module_parts = ['embedding_net', 'input_gate', 'dynamic_gates', 'head']

    def __init__(self, cfg: Config):
        super(EALSTM, self).__init__(cfg=cfg)
        self._hidden_size = cfg.hidden_size

        self.embedding_net = InputLayer(cfg)

        self.input_gate = nn.Linear(self.embedding_net.statics_output_size, cfg.hidden_size)

        # create tensors of learnable parameters
        self.dynamic_gates = _DynamicGates(cfg=cfg, input_size=self.embedding_net.dynamics_output_size)
        self.dropout = nn.Dropout(p=cfg.output_dropout)

        self.head = get_head(cfg=cfg, n_in=cfg.hidden_size, n_out=self.output_size)

    def _cell(self, x: torch.Tensor, i: torch.Tensor, states: Tuple[torch.Tensor,
                                                                    torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor]:
        """Single time step logic of EA-LSTM cell"""
        h_0, c_0 = states

        # calculate gates
        gates = self.dynamic_gates(h_0, x)
        f, o, g = gates.chunk(3, 1)

        c_1 = torch.sigmoid(f) * c_0 + i * torch.tanh(g)
        h_1 = torch.sigmoid(o) * torch.tanh(c_1)

        return h_1, c_1

    def forward(self, data: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform a forward pass on the EA-LSTM model.

        Parameters
        ----------
        data : Dict[str, torch.Tensor]
            Dictionary, containing input features as key-value pairs.

        Returns
        -------
        Dict[str, torch.Tensor]
            Model outputs and intermediate states as a dictionary. 
                - `y_hat`: model predictions of shape [batch size, sequence length, number of target variables].
                - `h_n`: hidden state at the last time step of the sequence of shape 
                    [batch size, sequence length, number of target variables].
                - `c_n`: cell state at the last time step of the sequence of shape 
                    [batch size, sequence length, number of target variables].
        """
        # possibly pass dynamic and static inputs through embedding layers
        x_d, x_s = self.embedding_net(data, concatenate_output=False)
        if x_s is None:
            raise ValueError('Need x_s or x_one_hot in forward pass.')

        # TODO: move hidden and cell state initialization to init and only reset states in forward pass to zero.
        h_t = x_d.data.new(x_d.shape[1], self._hidden_size).zero_()
        c_t = x_d.data.new(x_d.shape[1], self._hidden_size).zero_()

        # empty lists to temporally store all intermediate hidden/cell states
        h_n, c_n = [], []

        # calculate input gate only once because inputs are static
        i = torch.sigmoid(self.input_gate(x_s))

        # perform forward steps over input sequence
        for x_dt in x_d:

            h_t, c_t = self._cell(x_dt, i, (h_t, c_t))

            # store intermediate hidden/cell state in list
            h_n.append(h_t)
            c_n.append(c_t)

        h_n = torch.stack(h_n, 0).transpose(0, 1)
        c_n = torch.stack(c_n, 0).transpose(0, 1)

        pred = {'h_n': h_n, 'c_n': c_n}
        pred.update(self.head(self.dropout(h_n)))

        return pred


class _DynamicGates(nn.Module):
    """Internal class to wrap the dynamic gate parameters into a dedicated PyTorch Module"""

    def __init__(self, cfg: Config, input_size: int):
        super(_DynamicGates, self).__init__()
        self.cfg = cfg
        self.weight_ih = nn.Parameter(torch.FloatTensor(input_size, 3 * cfg.hidden_size))
        self.weight_hh = nn.Parameter(torch.FloatTensor(cfg.hidden_size, 3 * cfg.hidden_size))
        self.bias = nn.Parameter(torch.FloatTensor(3 * cfg.hidden_size))

        # initialize parameters
        self._reset_parameters()

    def _reset_parameters(self):
        """Special initialization of certain model weights."""
        nn.init.orthogonal_(self.weight_ih.data)

        weight_hh_data = torch.eye(self.cfg.hidden_size)
        weight_hh_data = weight_hh_data.repeat(1, 3)
        self.weight_hh.data = weight_hh_data

        nn.init.constant_(self.bias.data, val=0)

        if self.cfg.initial_forget_bias is not None:
            self.bias.data[:self.cfg.hidden_size] = self.cfg.initial_forget_bias

    def forward(self, h: torch.Tensor, x_d: torch.Tensor):
        gates = h @ self.weight_hh + x_d @ self.weight_ih + self.bias
        return gates



import logging
import subprocess
import sys
from pathlib import Path
from typing import Optional

LOGGER = logging.getLogger(__name__)


def setup_logging(log_file: str):
    """Initialize logging to `log_file` and stdout.

    Parameters
    ----------
    log_file : str
        Name of the file that will be logged to.
    """
    file_handler = logging.FileHandler(filename=log_file)
    stdout_handler = logging.StreamHandler(sys.stdout)

    logging.basicConfig(handlers=[file_handler, stdout_handler], level=logging.INFO, format="%(asctime)s: %(message)s")

    # Make sure we log uncaught exceptions
    def exception_logging(type, value, tb):
        LOGGER.exception(f"Uncaught exception", exc_info=(type, value, tb))

    sys.excepthook = exception_logging

    LOGGER.info(f"Logging to {log_file} initialized.")


def get_git_hash() -> Optional[str]:
    """Get git commit hash of the project if it is a git repository.

    Returns
    -------
    Optional[str]
        Git commit hash if project is a git repository, else None.
    """
    # get git commit hash if folder is a git repository
    current_dir = str(Path(__file__).absolute().parent)
    try:
        if subprocess.call(["git", "-C", current_dir, "branch"], stderr=subprocess.DEVNULL,
                           stdout=subprocess.DEVNULL) == 0:
            return subprocess.check_output(["git", "-C", current_dir, "describe", "--always"]).strip().decode('ascii')
    except OSError:
        return None  # likely, git is not installed.


def save_git_diff(run_dir: Path):
    """Try to store the git diff to a file.

    Parameters
    ----------
    run_dir : Path
        Directory of the current run.
    """
    base_dir = str(Path(__file__).absolute().parent)
    try:
        # diff should include staged and unstaged changes, hence we use "HEAD"
        out = subprocess.check_output(['git', '-C', base_dir, 'diff', 'HEAD'], stderr=subprocess.DEVNULL)
    except OSError:
        LOGGER.warning('Could not store git diff, likely because git is not installed '
                       'or because your version of git is too old (< 1.8.5)')
        return

    new_diff = out.strip().decode('utf-8')

    if new_diff:
        existing_diffs = list(run_dir.glob('neuralhydrology*.diff'))
        if len(existing_diffs) > 0:
            last_diff_path = run_dir / f'neuralhydrology-{len(existing_diffs) - 1}.diff'
            with last_diff_path.open('r') as last_diff_file:
                last_diff = last_diff_file.read()
            if last_diff == new_diff:
                LOGGER.info(f'Git repository contains uncommitted changes that are stored in {last_diff_path}.')
                return

        file_path = run_dir / f'neuralhydrology-{len(existing_diffs)}.diff'
        LOGGER.warning(f'Git repository contains uncommitted changes. Writing diff to {file_path}.')
        with file_path.open('w') as diff_file:
            diff_file.write(new_diff)



import torch

from neuralhydrology.utils.config import Config
from neuralhydrology.modelzoo.mclstm import MCLSTM


def test_mass_conservation():
    torch.manual_seed(111)

    # create minimal config required for model initialization
    config = Config({
        'dynamic_inputs': ['tmin(C)', 'tmax(C)'],
        'hidden_size': 10,
        'initial_forget_bias': 0,
        'mass_inputs': ['prcp(mm/day)'],
        'model': 'mclstm',
        'target_variables': ['QObs(mm/d)']
    })
    model = MCLSTM(config)

    # create random inputs
    data = {
        'x_d':
            torch.rand((1, 25, 3))  # [batch size, sequence length, total number of time series inputs]
    }

    # get model outputs and intermediate states
    output = model(data)

    # the total mass within the system at each time step is the cumsum over the outgoing mass + the current cell state
    cumsum_system = output["m_out"].sum(-1).cumsum(-1) + output["c"].sum(-1)

    # the accumulated mass of the inputs at each time step
    cumsum_input = data["x_d"][:, :, 0].cumsum(-1)

    # check if the total mass is conserved at every timestep of the forward pass
    assert torch.allclose(cumsum_system, cumsum_input)



import warnings
from datetime import datetime
from typing import Dict, List, Tuple

from dateutil.relativedelta import relativedelta

import numpy as np
from numba import njit
from xarray.core.dataarray import DataArray

from neuralhydrology.datautils import utils


def get_available_signatures() -> List[str]:
    """Return a list of available signatures.

    Returns
    -------
    List[str]
        List of all available signatures.
    """
    signatures = [
        "high_q_freq", "high_q_dur", "low_q_freq", "low_q_dur", "zero_q_freq", "q95", "q5", "q_mean", "hfd_mean",
        "baseflow_index", "slope_fdc", "stream_elas", "runoff_ratio"
    ]
    return signatures


def calculate_all_signatures(da: DataArray, prcp: DataArray, datetime_coord: str = None) -> Dict[str, float]:
    """Calculate all signatures with default values.

    Parameters
    ----------
    da : DataArray
        Array of discharge values for which the signatures will be calculated.
    prcp : DataArray
        Array of precipitation values.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified.

    Returns
    -------
    Dict[str, float]
        Dictionary with signature names as keys and signature values as values.
    """
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(da)

    results = {
        "high_q_freq": high_q_freq(da, datetime_coord=datetime_coord),
        "high_q_dur": high_q_dur(da),
        "low_q_freq": low_q_freq(da, datetime_coord=datetime_coord),
        "low_q_dur": low_q_dur(da),
        "zero_q_freq": zero_q_freq(da),
        "q95": q95(da),
        "q5": q5(da),
        "q_mean": q_mean(da),
        "hfd_mean": hfd_mean(da, datetime_coord=datetime_coord),
        "baseflow_index": baseflow_index(da, datetime_coord=datetime_coord)[0],
        "slope_fdc": slope_fdc(da),
        "stream_elas": stream_elas(da, prcp, datetime_coord=datetime_coord),
        "runoff_ratio": runoff_ratio(da, prcp, datetime_coord=datetime_coord)
    }
    return results


def calculate_signatures(da: DataArray,
                         signatures: List[str],
                         datetime_coord: str = None,
                         prcp: DataArray = None) -> Dict[str, float]:
    """Calculate the specified signatures with default values.

    Parameters
    ----------
    da : DataArray
        Array of discharge values for which the signatures will be calculated.
    signatures : List[str]
        List of names of the signatures to calculate.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified.
    prcp : DataArray, optional
        Array of precipitation values. Required for signatures 'runoff_ratio' and 'streamflow_elas'.

    Returns
    -------
    Dict[str, float]
        Dictionary with signature names as keys and signature values as values.

    Raises
    ------
    ValueError
        If a passed signature name does not exist.
    """
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(da)

    values = {}
    for signature in signatures:
        if signature == "high_q_freq":
            values["high_q_freq"] = high_q_freq(da, datetime_coord=datetime_coord)
        elif signature == "high_q_dur":
            values["high_q_dur"] = high_q_dur(da)
        elif signature == "low_q_freq":
            values["low_q_freq"] = low_q_freq(da, datetime_coord=datetime_coord)
        elif signature == "low_q_dur":
            values["low_q_dur"] = low_q_dur(da)
        elif signature == "zero_q_freq":
            values["zero_q_freq"] = zero_q_freq(da)
        elif signature == "q95":
            values["q95"] = q95(da)
        elif signature == "q5":
            values["q5"] = q5(da)
        elif signature == "q_mean":
            values["q_mean"] = q_mean(da)
        elif signature == "hfd_mean":
            values["hfd_mean"] = hfd_mean(da, datetime_coord=datetime_coord)
        elif signature == "baseflow_index":
            values["baseflow_index"] = baseflow_index(da, datetime_coord=datetime_coord)[0]
        elif signature == "slope_fdc":
            values["slope_fdc"] = slope_fdc(da)
        elif signature == "runoff_ratio":
            values["runoff_ratio"] = runoff_ratio(da, prcp, datetime_coord=datetime_coord)
        elif signature == "stream_elas":
            values["stream_elas"] = stream_elas(da, prcp, datetime_coord=datetime_coord)
        else:
            ValueError(f"Unknown signatures {signature}")
    return values


@njit
def _split_list(alist: list, min_length: int = 0) -> list:
    """Split a list of indices into lists of consecutive indices of at least length `min_length`. """
    newlist = []
    start = 0
    for index, value in enumerate(alist):
        if index < len(alist) - 1:
            if alist[index + 1] > value + 1:
                end = index + 1
                if end - start >= min_length:
                    newlist.append(alist[start:end])
                start = end
        else:
            if len(alist) - start >= min_length:
                newlist.append(alist[start:len(alist)])
    return newlist


def high_q_dur(da: DataArray, threshold: float = 9.) -> float:
    """Calculate high-flow duration.

    Average duration of high-flow events (number of consecutive steps >`threshold` times the median flow) [#]_,
    [#]_ (Table 2).

    Parameters
    ----------
    da : DataArray
        Array of flow values.
    threshold : float, optional
        High-flow threshold. Values larger than ``threshold * median`` are considered high flows.

    Returns
    -------
    float
        High-flow duration

    References
    ----------
    .. [#] Clausen, B. and Biggs, B. J. F.: Flow variables for ecological studies in temperate streams: groupings based
        on covariance. Journal of Hydrology, 2000, 237, 184--197, doi:10.1016/S0022-1694(00)00306-1
    .. [#] Westerberg, I. K. and McMillan, H. K.: Uncertainty in hydrological signatures.
        Hydrology and Earth System Sciences, 2015, 19, 3951--3968, doi:10.5194/hess-19-3951-2015
    """
    median_flow = float(da.median())
    idx = np.where(da.values > threshold * median_flow)[0]
    if len(idx) > 0:
        periods = _split_list(idx)
        hqd = np.mean([len(p) for p in periods])
    else:
        hqd = np.nan
    return hqd


def low_q_dur(da: DataArray, threshold: float = 0.2) -> float:
    """Calculate low-flow duration.

    Average duration of low-flow events (number of consecutive steps <`threshold` times the median flow) [#]_,
    [#]_ (Table 2).

    Parameters
    ----------
    da : DataArray
        Array of flow values.
    threshold : float, optional
        Low-flow threshold. Values below ``threshold * median`` are considered low flows.

    Returns
    -------
    float
        Low-flow duration

    References
    ----------
    .. [#] Olden, J. D. and Poff, N. L.: Redundancy and the choice of hydrologic indices for characterizing streamflow
        regimes. River Research and Applications, 2003, 19, 101--121, doi:10.1002/rra.700
    .. [#] Westerberg, I. K. and McMillan, H. K.: Uncertainty in hydrological signatures.
        Hydrology and Earth System Sciences, 2015, 19, 3951--3968, doi:10.5194/hess-19-3951-2015
    """
    mean_flow = float(da.mean())
    idx = np.where(da.values < threshold * mean_flow)[0]
    if len(idx) > 0:
        periods = _split_list(idx)
        lqd = np.mean([len(p) for p in periods])
    else:
        lqd = np.nan
    return lqd


def zero_q_freq(da: DataArray) -> float:
    """Calculate zero-flow frequency.

    Frequency of steps with zero discharge.

    Parameters
    ----------
    da : DataArray
        Array of flow values.

    Returns
    -------
    float
        Zero-flow frequency.
    """
    # number of steps with zero flow
    n_steps = (da == 0).sum()

    return float(n_steps / len(da))


def high_q_freq(da: DataArray, datetime_coord: str = None, threshold: float = 9.) -> float:
    """Calculate high-flow frequency.

    Frequency of high-flow events (>`threshold` times the median flow) [#]_, [#]_ (Table 2).

    Parameters
    ----------
    da : DataArray
        Array of flow values.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified.
    threshold : float, optional
        High-flow threshold. Values larger than ``threshold * median`` are considered high flows.

    Returns
    -------
    float
        High-flow frequency

    References
    ----------
    .. [#] Clausen, B. and Biggs, B. J. F.: Flow variables for ecological studies in temperate streams: groupings based
        on covariance. Journal of Hydrology, 2000, 237, 184--197, doi:10.1016/S0022-1694(00)00306-1
    .. [#] Westerberg, I. K. and McMillan, H. K.: Uncertainty in hydrological signatures.
        Hydrology and Earth System Sciences, 2015, 19, 3951--3968, doi:10.5194/hess-19-3951-2015
    """
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(da)

    # determine the date of the first January 1st in the data period
    first_date = da.coords[datetime_coord][0].values.astype('datetime64[s]').astype(datetime)
    last_date = da.coords[datetime_coord][-1].values.astype('datetime64[s]').astype(datetime)

    if first_date == datetime.strptime(f'{first_date.year}-01-01', '%Y-%m-%d'):
        start_date = first_date
    else:
        start_date = datetime.strptime(f'{first_date.year + 1}-01-01', '%Y-%m-%d')

    # end date of the first full year period
    end_date = start_date + relativedelta(years=1) - relativedelta(seconds=1)

    # determine the median flow over the entire period
    median_flow = da.median(skipna=True)

    hqfs = []
    while end_date < last_date:

        data = da.sel({datetime_coord: slice(start_date, end_date)})

        # number of steps with discharge higher than threshold * median in a one year period
        n_steps = (data > (threshold * median_flow)).sum()

        hqfs.append(float(n_steps))

        start_date += relativedelta(years=1)
        end_date += relativedelta(years=1)

    return np.mean(hqfs)


def low_q_freq(da: DataArray, datetime_coord: str = None, threshold: float = 0.2) -> float:
    """Calculate Low-flow frequency.

    Frequency of low-flow events (<`threshold` times the median flow) [#]_, [#]_ (Table 2).

    Parameters
    ----------
    da : DataArray
        Array of flow values.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified.
    threshold : float, optional
        Low-flow threshold. Values below ``threshold * median`` are considered low flows.

    Returns
    -------
    float
        Low-flow frequency

    References
    ----------
    .. [#] Olden, J. D. and Poff, N. L.: Redundancy and the choice of hydrologic indices for characterizing streamflow
        regimes. River Research and Applications, 2003, 19, 101--121, doi:10.1002/rra.700
    .. [#] Westerberg, I. K. and McMillan, H. K.: Uncertainty in hydrological signatures.
        Hydrology and Earth System Sciences, 2015, 19, 3951--3968, doi:10.5194/hess-19-3951-2015
    """
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(da)

    # determine the date of the first January 1st in the data period
    first_date = da.coords[datetime_coord][0].values.astype('datetime64[s]').astype(datetime)
    last_date = da.coords[datetime_coord][-1].values.astype('datetime64[s]').astype(datetime)

    if first_date == datetime.strptime(f'{first_date.year}-01-01', '%Y-%m-%d'):
        start_date = first_date
    else:
        start_date = datetime.strptime(f'{first_date.year + 1}-01-01', '%Y-%m-%d')

    # end date of the first full year period
    end_date = start_date + relativedelta(years=1) - relativedelta(seconds=1)

    # determine the mean flow over the entire period
    mean_flow = da.mean(skipna=True)

    lqfs = []
    while end_date < last_date:

        data = da.sel({datetime_coord: slice(start_date, end_date)})

        # number of steps with discharge lower than threshold * median in a one year period
        n_steps = (data < (threshold * mean_flow)).sum()

        lqfs.append(float(n_steps))

        start_date += relativedelta(years=1)
        end_date += relativedelta(years=1)

    return np.mean(lqfs)


def hfd_mean(da: DataArray, datetime_coord: str = None) -> float:
    """Calculate mean half-flow duration.

    Mean half-flow date (step on which the cumulative discharge since October 1st
    reaches half of the annual discharge) [#]_.

    Parameters
    ----------
    da : DataArray
        Array of flow values.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified.

    Returns
    -------
    float
        Mean half-flow duration

    References
    ----------
    .. [#] Court, A.: Measures of streamflow timing. Journal of Geophysical Research (1896-1977), 1962, 67, 4335--4339,
        doi:10.1029/JZ067i011p04335
    """
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(da)

    # determine the date of the first October 1st in the data period
    first_date = da.coords[datetime_coord][0].values.astype('datetime64[s]').astype(datetime)
    last_date = da.coords[datetime_coord][-1].values.astype('datetime64[s]').astype(datetime)

    if first_date > datetime.strptime(f'{first_date.year}-10-01', '%Y-%m-%d'):
        start_date = datetime.strptime(f'{first_date.year + 1}-10-01', '%Y-%m-%d')
    else:
        start_date = datetime.strptime(f'{first_date.year}-10-01', '%Y-%m-%d')

    end_date = start_date + relativedelta(years=1) - relativedelta(seconds=1)

    doys = []
    while end_date < last_date:

        # compute cumulative sum for the selected period
        data = da.sel({datetime_coord: slice(start_date, end_date)})
        cs = data.cumsum(skipna=True)

        # find steps with more cumulative discharge than the half annual sum
        hf_steps = np.where(~np.isnan(cs.where(cs > data.sum(skipna=True) / 2).values))[0]

        # ignore days without discharge
        if len(hf_steps) > 0:
            # store the first step in the result array
            doys.append(hf_steps[0])

        start_date += relativedelta(years=1)
        end_date += relativedelta(years=1)

    return np.mean(doys)


def q5(da: DataArray) -> float:
    """Calculate 5th flow quantile.

    Parameters
    ----------
    da : DataArray
        Array of flow values.

    Returns
    -------
    float
        5th flow quantile.
    """
    return float(da.quantile(0.05))


def q95(da: DataArray) -> float:
    """Calculate 95th flow quantile.

    Parameters
    ----------
    da : DataArray
        Array of flow values.

    Returns
    -------
    float
        95th flow quantile.
    """
    return float(da.quantile(0.95))


def q_mean(da: DataArray) -> float:
    """Calculate mean discharge.

    Parameters
    ----------
    da : DataArray
        Array of flow values.

    Returns
    -------
    float
        Mean discharge.
    """
    return float(da.mean())


@njit
def _baseflow_index_jit(streamflow: np.ndarray, alpha: float, warmup: int, n_passes: int) -> Tuple[float, np.ndarray]:
    non_nan_indices = np.where(~np.isnan(streamflow))[0]
    non_nan_streamflow_runs = _split_list(non_nan_indices, min_length=warmup + 1)
    if len(non_nan_streamflow_runs) == 0:
        # no consecutive run of non-NaN values is long enough to calculate baseflow.
        return np.nan, np.full_like(streamflow, np.nan)

    streamflow_sum = 0
    overall_baseflow = np.full_like(streamflow, np.nan)
    for non_nan_run in non_nan_streamflow_runs:
        streamflow_run = streamflow[non_nan_run]

        # mirror discharge of length 'window' at the start and end
        padded_streamflow = np.zeros((len(streamflow_run) + 2 * warmup))
        padded_streamflow[warmup:-warmup] = streamflow_run
        padded_streamflow[:warmup] = streamflow_run[1:warmup + 1][::-1]
        padded_streamflow[-warmup:] = streamflow_run[-warmup - 1:-1][::-1]

        baseflow = padded_streamflow
        for _ in range(n_passes):
            new_baseflow = np.zeros_like(padded_streamflow)
            quickflow = baseflow[0]
            for i in range(1, len(padded_streamflow)):
                quickflow = alpha * quickflow + (1 + alpha) * (baseflow[i] - baseflow[i - 1]) / 2
                if quickflow > 0:
                    new_baseflow[i] = baseflow[i] - quickflow
                else:
                    new_baseflow[i] = baseflow[i]

            # switch between forward and backward passes. Next iteration's input is the baseflow generated in this iteration
            baseflow = np.flip(new_baseflow)

        overall_baseflow[non_nan_run] = baseflow[warmup:-warmup]
        streamflow_sum += np.sum(streamflow_run)

    bf_index = np.nansum(overall_baseflow) / streamflow_sum

    return bf_index, overall_baseflow


def baseflow_index(da: DataArray,
                   alpha: float = 0.98,
                   warmup: int = 30,
                   n_passes: int = None,
                   datetime_coord: str = None) -> Tuple[float, DataArray]:
    """Calculate baseflow index.

    Ratio of mean baseflow to mean discharge [#]_. If `da` contains NaN values, the baseflow is calculated for each
    consecutive segment of more than `warmup` non-NaN values.

    Parameters
    ----------
    da : DataArray
        Array of flow values.
    alpha : float, optional
        alpha filter parameter.
    warmup : int, optional
        Number of warmup steps.
    n_passes : int, optional
        Number of passes (alternating forward and backward) to perform. Should be an odd number. If None, will use
        3 for daily and 9 for hourly data and fail for all other input frequencies.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified. Used to infer the 
        frequency if `n_passes` is None.

    Returns
    -------
    Tuple[float, DataArray]
        Baseflow index and baseflow array. The baseflow array contains NaNs wherever no baseflow was
        calculated due to NaNs in `da`.

    Raises
    ------
    ValueError
        If `da` has a frequency other than daily or hourly and `n_passes` is None.

    References
    ----------
    .. [#] Ladson, T. R., Brown, R., Neal, B., and Nathan, R.: A Standard Approach to Baseflow Separation Using The
        Lyne and Hollick Filter. Australasian Journal of Water Resources, Taylor & Francis, 2013, 17, 25--34,
        doi:10.7158/13241583.2013.11465417
    """
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(da)

    if n_passes is None:
        freq = utils.infer_frequency(da[datetime_coord].values)
        if freq == '1D':
            n_passes = 3
        elif freq == '1H':
            n_passes = 9
        else:
            raise ValueError(f'For frequencies other than daily or hourly, n_passes must be specified.')
    if n_passes % 2 != 1:
        warnings.warn('n_passes should be an even number. The returned baseflow will be reversed.')

    # call jit compiled function to calculate baseflow
    bf_index, baseflow = _baseflow_index_jit(da.values, alpha, warmup, n_passes)

    # parse baseflow as a DataArray using the coordinates of the streamflow array
    da_baseflow = da.copy()
    da_baseflow.data = baseflow

    return bf_index, da_baseflow


def slope_fdc(da: DataArray, lower_quantile: float = 0.33, upper_quantile: float = 0.66) -> float:
    """Calculates flow duration curve slope.

    Slope of the flow duration curve (between the log-transformed `lower_quantile` and `upper_quantile`) [#]_ (Eq. 3).

    Parameters
    ----------
    da : DataArray
        Array of flow values.
    lower_quantile : float, optional
        Lower quantile to use in slope calculation.
    upper_quantile : float, optional
        Upper quantile to use in slope calculation.

    Returns
    -------
    float
        Slope of the flow duration curve.

    References
    ----------
    .. [#] Sawicz, K., Wagener, T., Sivapalan, M., Troch, P. A., and Carrillo, G.: Catchment classification: empirical
        analysis of hydrologic similarity based on catchment function in the eastern USA.
        Hydrology and Earth System Sciences, 2011, 15, 2895--2911, doi:10.5194/hess-15-2895-2011
    """
    # sort discharge by descending order
    fdc = da.sortby(da, ascending=False)

    # get idx of lower and upper quantile
    idx_lower = np.round(lower_quantile * len(fdc)).astype(int)
    idx_upper = np.round(upper_quantile * len(fdc)).astype(int)

    value = (np.log(fdc[idx_lower].values + 1e-8) - np.log(fdc[idx_upper].values + 1e-8)) / (upper_quantile -
                                                                                             lower_quantile)

    return value


def runoff_ratio(da: DataArray, prcp: DataArray, datetime_coord: str = None) -> float:
    """Calculate runoff ratio.

    Runoff ratio (ratio of mean discharge to mean precipitation) [#]_ (Eq. 2).

    Parameters
    ----------
    da : DataArray
        Array of flow values.
    prcp : DataArray
        Array of precipitation values.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified.

    Returns
    -------
    float
        Runoff ratio.

    References
    ----------
    .. [#] Sawicz, K., Wagener, T., Sivapalan, M., Troch, P. A., and Carrillo, G.: Catchment classification: empirical
        analysis of hydrologic similarity based on catchment function in the eastern USA.
        Hydrology and Earth System Sciences, 2011, 15, 2895--2911, doi:10.5194/hess-15-2895-2011
    """
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(da)

    # rename precip coordinate name (to avoid problems with 'index' or 'date')
    prcp = prcp.rename({list(prcp.coords.keys())[0]: datetime_coord})

    # slice prcp to the same time window as the discharge
    prcp = prcp.sel({datetime_coord: slice(da.coords[datetime_coord][0], da.coords[datetime_coord][-1])})

    # calculate runoff ratio
    value = da.mean() / prcp.mean()

    return float(value)


def stream_elas(da: DataArray, prcp: DataArray, datetime_coord: str = None) -> float:
    """Calculate stream elasticity.

    Streamflow precipitation elasticity (sensitivity of streamflow to changes in precipitation at
    the annual time scale) [#]_.

    Parameters
    ----------
    da : DataArray
        Array of flow values.
    prcp : DataArray
        Array of precipitation values.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified.

    Returns
    -------
    float
        Stream elasticity.

    References
    ----------
    .. [#] Sankarasubramanian, A., Vogel, R. M., and Limbrunner, J. F.: Climate elasticity of streamflow in the
        United States. Water Resources Research, 2001, 37, 1771--1781, doi:10.1029/2000WR900330
    """
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(da)

    # rename precip coordinate name (to avoid problems with 'index' or 'date')
    prcp = prcp.rename({list(prcp.coords.keys())[0]: datetime_coord})

    # slice prcp to the same time window as the discharge
    prcp = prcp.sel({datetime_coord: slice(da.coords[datetime_coord][0], da.coords[datetime_coord][-1])})

    # determine the date of the first October 1st in the data period
    first_date = da.coords[datetime_coord][0].values.astype('datetime64[s]').astype(datetime)
    last_date = da.coords[datetime_coord][-1].values.astype('datetime64[s]').astype(datetime)

    if first_date > datetime.strptime(f'{first_date.year}-10-01', '%Y-%m-%d'):
        start_date = datetime.strptime(f'{first_date.year + 1}-10-01', '%Y-%m-%d')
    else:
        start_date = datetime.strptime(f'{first_date.year}-10-01', '%Y-%m-%d')

    end_date = start_date + relativedelta(years=1) - relativedelta(seconds=1)

    # mask only valid time steps (only discharge has missing values)
    idx = (da >= 0) & (~da.isnull())
    da = da[idx]
    prcp = prcp[idx]

    # calculate long-term means
    q_mean_total = da.mean()
    p_mean_total = prcp.mean()

    values = []
    while end_date < last_date:
        q = da.sel({datetime_coord: slice(start_date, end_date)})
        p = prcp.sel({datetime_coord: slice(start_date, end_date)})

        val = (q.mean() - q_mean_total) / (p.mean() - p_mean_total) * (p_mean_total / q_mean_total)
        values.append(val)

        start_date += relativedelta(years=1)
        end_date += relativedelta(years=1)

    return np.median([float(v) for v in values])



import math
from collections import defaultdict

from typing import Dict, Union

import torch
import torch.nn as nn

from neuralhydrology.modelzoo.basemodel import BaseModel
from neuralhydrology.modelzoo.cudalstm import CudaLSTM
from neuralhydrology.modelzoo.embcudalstm import EmbCudaLSTM
from neuralhydrology.modelzoo.inputlayer import InputLayer
from neuralhydrology.modelzoo.head import get_head
from neuralhydrology.utils.config import Config


class CustomLSTM(BaseModel):
    """A custom implementation of the LSTM with support for a dedicated embedding network.

    The idea of this model is mainly to be used as an analytical tool, where you can train a model using the optimized
    `CudaLSTM` or `EmbCudaLSTM` classes, and later copy the weights into this model for a more in-depth network
    analysis (e.g. inspecting model states or gate activations). The advantage of this implementation is that it returns
    the entire time series of state vectors and gate activations.
    However, you can also use this model class for training but note that it will be considerably slower than its
    optimized counterparts. Depending on the embedding settings, static and/or dynamic features may or may not be fed
    through embedding networks before being concatenated and passed through the LSTM.

    Parameters
    ----------
    cfg : Config
        The run configuration.

    Example
    -------
    >>> # Example for copying the weights of an optimzed `CudaLSTM` or `EmbCudaLSTM` into a `CustomLSTM` instance
    >>> cfg = ... # A config instance corresponding to the original, optimized model
    >>> optimized_lstm = ... # A model instance of `CudaLSTM` or `EmbCudaLSTM`
    >>>
    >>> # Use the original config to initialize this model to differentiate between `CudaLSTM` and `EmbCudaLSTM`
    >>> custom_lstm = CustomLSTM(cfg=cfg)
    >>>
    >>> # Copy weights into the `LSTM` instance.
    >>> custom_lstm.copy_weights(optimized_lstm)
    """

    def __init__(self, cfg: Config):
        super(CustomLSTM, self).__init__(cfg=cfg)

        self.embedding_net = InputLayer(cfg)

        self._hidden_size = cfg.hidden_size

        self.cell = _LSTMCell(input_size=self.embedding_net.output_size,
                              hidden_size=self._hidden_size,
                              initial_forget_bias=cfg.initial_forget_bias)

        self.dropout = nn.Dropout(p=cfg.output_dropout)

        self.head = get_head(cfg=cfg, n_in=self._hidden_size, n_out=self.output_size)

    def forward(self,
                data: Dict[str, torch.Tensor],
                h_0: torch.Tensor = None,
                c_0: torch.Tensor = None) -> Dict[str, torch.Tensor]:
        """Perform a forward pass on the LSTM model.

        Parameters
        ----------
        data : Dict[str, torch.Tensor]
            Dictionary, containing input features as key-value pair.
        h_0 : torch.Tensor, optional
            Initial hidden state, by default 0.
        c_0 : torch.Tensor, optional
            Initial cell state, by default 0.

        Returns
        -------
        Dict[str, torch.Tensor]
            Model output and all intermediate states and gate activations as a dictionary.
        """
        # possibly pass dynamic and static inputs through embedding layers, then concatenate them
        x_d = self.embedding_net(data, concatenate_output=True)

        seq_len, batch_size, _ = x_d.size()

        # TODO: move hidden and cell state initialization to init and only reset states in forward pass to zero.
        if h_0 is None:
            h_0 = x_d.data.new(batch_size, self._hidden_size).zero_()
        if c_0 is None:
            c_0 = x_d.data.new(batch_size, self._hidden_size).zero_()
        h_x = (h_0, c_0)

        output = defaultdict(list)
        for x_t in x_d:
            h_0, c_0 = h_x
            cell_output = self.cell(x_t=x_t, h_0=h_0, c_0=c_0)

            h_x = (cell_output['h_n'], cell_output['c_n'])

            for key, cell_out in cell_output.items():
                output[key].append(cell_out)

        # stack to [batch size, sequence length, hidden size]
        pred = {key: torch.stack(val, 1) for key, val in output.items()}
        pred.update(self.head(self.dropout(pred['h_n'])))
        return pred

    def copy_weights(self, optimized_lstm: Union[CudaLSTM, EmbCudaLSTM]):
        """Copy weights from a `CudaLSTM` or `EmbCudaLSTM` into this model class

        Parameters
        ----------
        optimized_lstm : Union[CudaLSTM, EmbCudaLSTM]
            Model instance of a `CudaLSTM` (neuralhydrology.modelzoo.cudalstm) or `EmbCudaLSTM`
            (neuralhydrology.modelzoo.embcudalstm).
            
        Raises
        ------
        RuntimeError
            If `optimized_lstm` is an `EmbCudaLSTM` but this model instance was not created with an embedding network.
        """

        assert isinstance(optimized_lstm, CudaLSTM) or isinstance(optimized_lstm, EmbCudaLSTM)

        self.embedding_net.load_state_dict(optimized_lstm.embedding_net.state_dict())

        # copy lstm cell weights
        self.cell.copy_weights(optimized_lstm.lstm, layer=0)

        # copy weights of head
        self.head.load_state_dict(optimized_lstm.head.state_dict())


class _LSTMCell(nn.Module):

    def __init__(self, input_size: int, hidden_size: int, initial_forget_bias: float = 0.0):
        super(_LSTMCell, self).__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.initial_forget_bias = initial_forget_bias

        self.w_hh = nn.Parameter(torch.FloatTensor(4 * hidden_size, hidden_size))
        self.w_ih = nn.Parameter(torch.FloatTensor(4 * hidden_size, input_size))

        self.b_hh = nn.Parameter(torch.FloatTensor(4 * hidden_size))
        self.b_ih = nn.Parameter(torch.FloatTensor(4 * hidden_size))

        self._reset_parameters()

    def _reset_parameters(self):
        """Special initialization of certain model weights."""
        stdv = math.sqrt(3 / self.hidden_size)
        for weight in self.parameters():
            if len(weight.shape) > 1:
                weight.data.uniform_(-stdv, stdv)
            else:
                nn.init.zeros_(weight)

        if self.initial_forget_bias != 0:
            self.b_hh.data[self.hidden_size:2 * self.hidden_size] = self.initial_forget_bias

    def forward(self, x_t: torch.Tensor, h_0: torch.Tensor, c_0: torch.Tensor) -> Dict[str, torch.Tensor]:
        gates = h_0 @ self.w_hh.T + self.b_hh + x_t @ self.w_ih.T + self.b_ih
        i, f, g, o = gates.chunk(4, 1)

        c_1 = c_0 * torch.sigmoid(f) + torch.sigmoid(i) * torch.tanh(g)
        h_1 = torch.sigmoid(o) * torch.tanh(c_1)

        return {'h_n': h_1, 'c_n': c_1, 'i': i, 'f': f, 'g': g, 'o': o}

    def copy_weights(self, cudnnlstm: nn.Module, layer: int):

        assert self.hidden_size == cudnnlstm.hidden_size
        assert self.input_size == cudnnlstm.input_size

        self.w_hh.data = getattr(cudnnlstm, f"weight_hh_l{layer}").data
        self.w_ih.data = getattr(cudnnlstm, f"weight_ih_l{layer}").data
        self.b_hh.data = getattr(cudnnlstm, f"bias_hh_l{layer}").data
        self.b_ih.data = getattr(cudnnlstm, f"bias_ih_l{layer}").data



import functools
import pickle
import re
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Union

import numpy as np
import pandas as pd
import xarray
from pandas.tseries.frequencies import to_offset
from ruamel.yaml import YAML
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset


def load_scaler(run_dir: Path) -> Dict[str, Union[pd.Series, xarray.Dataset]]:
    """Load feature scaler from run directory.

    Checks run directory for scaler file in yaml format (new) or pickle format (old).

    Parameters
    ----------
    run_dir: Path
        Run directory. Has to contain a folder 'train_data' that contains the 'train_data_scaler' file.

    Returns
    -------
    Dictionary, containing the feature scaler for static and dynamic features.
    
    Raises
    ------
    FileNotFoundError
        If neither a 'train_data_scaler.yml' or 'train_data_scaler.p' file is found in the 'train_data' folder of the 
        run directory.
    """
    scaler_file = run_dir / "train_data" / "train_data_scaler.yml"

    if scaler_file.is_file():
        # read scaler from disk
        with scaler_file.open("r") as fp:
            yaml = YAML(typ="safe")
            scaler_dump = yaml.load(fp)

        # transform scaler into the format expected by NeuralHydrology
        scaler = {}
        for key, value in scaler_dump.items():
            if key in ["attribute_means", "attribute_stds", "camels_attr_means", "camels_attr_stds"]:
                scaler[key] = pd.Series(value)
            elif key in ["xarray_feature_scale", "xarray_feature_center"]:
                scaler[key] = xarray.Dataset.from_dict(value).astype(np.float32)

        return scaler

    else:
        scaler_file = run_dir / "train_data" / "train_data_scaler.p"

        if scaler_file.is_file():
            with scaler_file.open('rb') as fp:
                scaler = pickle.load(fp)
            return scaler
        else:
            raise FileNotFoundError(f"No scaler file found in {scaler_file.parent}. "
                                    "Looked for (new) yaml file or (old) pickle file")


def load_hydroatlas_attributes(data_dir: Path, basins: List[str] = []) -> pd.DataFrame:
    """Load HydroATLAS attributes into a pandas DataFrame

    Parameters
    ----------
    data_dir : Path
        Path to the root directory of the dataset. Must contain a folder called 'hydroatlas_attributes' with a file
        called `attributes.csv`. The attributes file is expected to have one column called `basin_id`.
    basins : List[str], optional
        If passed, return only attributes for the basins specified in this list. Otherwise, the attributes of all basins
        are returned.

    Returns
    -------
    pd.DataFrame
        Basin-indexed DataFrame containing the HydroATLAS attributes.
    """
    attribute_file = data_dir / "hydroatlas_attributes" / "attributes.csv"
    if not attribute_file.is_file():
        raise FileNotFoundError(attribute_file)

    df = pd.read_csv(attribute_file, dtype={'basin_id': str})
    df = df.set_index('basin_id')

    if basins:
        drop_basins = [b for b in df.index if b not in basins]
        df = df.drop(drop_basins, axis=0)

    return df


def load_basin_file(basin_file: Path) -> List[str]:
    """Load list of basins from text file.
    
    Note: Basins names are not allowed to end with '_period*'
    
    Parameters
    ----------
    basin_file : Path
        Path to a basin txt file. File has to contain one basin id per row, while empty rows are ignored.

    Returns
    -------
    List[str]
        List of basin ids as strings.
        
    Raises
    ------
    ValueError
        In case of invalid basin names that would cause problems internally.
    """
    with basin_file.open('r') as fp:
        basins = sorted(basin.strip() for basin in fp if basin.strip())

    # sanity check basin names
    problematic_basins = [basin for basin in basins if basin.split('_')[-1].startswith('period')]
    if problematic_basins:
        msg = [
            f"The following basin names are invalid {problematic_basins}. Check documentation of the ",
            "'load_basin_file()' functions for details."
        ]
        raise ValueError(" ".join(msg))

    return basins


def attributes_sanity_check(df: pd.DataFrame):
    """Utility function to check the suitability of the attributes for model training.
    
    This utility function can be used to check if any attribute has a standard deviation of zero. This would lead to 
    NaN's when normalizing the features and thus would lead to NaN's when training the model. It also checks if any
    attribute for any basin contains a NaN, which would also cause NaNs during model training.
    
    Parameters
    ----------
    df : pd.DataFrame
        DataFrame of catchment attributes as columns.

    Raises
    ------
    RuntimeError
        If one or more attributes have a standard deviation of zero or any attribute for any basin is NaN.
    """
    # Check for NaNs in standard deviation of attributes.
    attributes = []
    if any(df.std() == 0.0) or any(df.std().isnull()):
        for k, v in df.std().iteritems():
            if (v == 0) or (np.isnan(v)):
                attributes.append(k)
    if attributes:
        msg = [
            "The following attributes have a std of zero or NaN, which results in NaN's ",
            "when normalizing the features. Remove the attributes from the attribute feature list ",
            "and restart the run. \n", f"Attributes: {attributes}"
        ]
        raise RuntimeError("".join(msg))

    # Check for NaNs in any attribute of any basin
    nan_df = df[df.isnull().any(axis=1)]
    if len(nan_df) > 0:
        failure_cases = defaultdict(list)
        for basin, row in nan_df.iterrows():
            for feature, value in row.iteritems():
                if np.isnan(value):
                    failure_cases[basin].append(feature)
        # create verbose error message
        msg = ["The following basins/attributes are NaN, which can't be used as input:"]
        for basin, features in failure_cases.items():
            msg.append(f"{basin}: {features}")
        raise RuntimeError("\n".join(msg))


def sort_frequencies(frequencies: List[str]) -> List[str]:
    """Sort the passed frequencies from low to high frequencies.

    Use `pandas frequency strings
    <https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases>`_
    to define frequencies. Note: The strings need to include values, e.g., '1D' instead of 'D'.

    Parameters
    ----------
    frequencies : List[str]
        List of pandas frequency identifiers to be sorted.

    Returns
    -------
    List[str]
        Sorted list of pandas frequency identifiers.

    Raises
    ------
    ValueError
        If a pair of frequencies in `frequencies` is not comparable via `compare_frequencies`.
    """
    return sorted(frequencies, key=functools.cmp_to_key(compare_frequencies))


def infer_frequency(index: Union[pd.DatetimeIndex, np.ndarray]) -> str:
    """Infer the frequency of an index of a pandas DataFrame/Series or xarray DataArray.

    Parameters
    ----------
    index : Union[pd.DatetimeIndex, np.ndarray]
        DatetimeIndex of a DataFrame/Series or array of datetime values.

    Returns
    -------
    str
        Frequency of the index as a `pandas frequency string
        <https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases>`_

    Raises
    ------
    ValueError
        If the frequency cannot be inferred from the index or is zero.
    """
    native_frequency = pd.infer_freq(index)
    if native_frequency is None:
        raise ValueError(f'Cannot infer a legal frequency from dataset: {native_frequency}.')
    if native_frequency[0] not in '0123456789':  # add a value to the unit so to_timedelta works
        native_frequency = f'1{native_frequency}'

    # pd.Timedelta doesn't understand weekly (W) frequencies, so we convert them to the equivalent multiple of 7D.
    weekly_freq = re.match('(\d+)W(-(MON|TUE|WED|THU|FRI|SAT|SUN))?$', native_frequency)
    if weekly_freq is not None:
        n = int(weekly_freq[1]) * 7
        native_frequency = f'{n}D'

    # Assert that the frequency corresponds to a positive time delta. We first add one offset to the base datetime
    # to make sure it's aligned with the frequency. Otherwise, adding an offset of e.g. 0Y would round up to the
    # nearest year-end, so we'd incorrectly miss a frequency of zero.
    base_datetime = pd.to_datetime('2001-01-01 00:00:00') + to_offset(native_frequency)
    if base_datetime >= base_datetime + to_offset(native_frequency):
        raise ValueError('Inferred dataset frequency is zero or negative.')
    return native_frequency


def infer_datetime_coord(xr: Union[DataArray, Dataset]) -> str:
    """Checks for coordinate with 'date' in its name and returns the name.
    
    Parameters
    ----------
    xr : Union[DataArray, Dataset]
        Array to infer coordinate name of.
        
    Returns
    -------
    str
        Name of datetime coordinate name.
        
    Raises
    ------
    RuntimeError
        If none or multiple coordinates with 'date' in its name are found.
    """
    candidates = [c for c in list(xr.coords) if "date" in c]
    if len(candidates) > 1:
        raise RuntimeError("Found multiple coordinates with 'date' in its name.")
    if not candidates:
        raise RuntimeError("Did not find any coordinate with 'date' in its name")

    return candidates[0]


def compare_frequencies(freq_one: str, freq_two: str) -> int:
    """Compare two frequencies.

    Note that only frequencies that work with `get_frequency_factor` can be compared.

    Parameters
    ----------
    freq_one : str
        First frequency.
    freq_two : str
        Second frequency.

    Returns
    -------
    int
        -1 if `freq_one` is lower than `freq_two`, +1 if it is larger, 0 if they are equal.

    Raises
    ------
    ValueError
        If the two frequencies are not comparable via `get_frequency_factor`.
    """
    freq_factor = get_frequency_factor(freq_one, freq_two)
    if freq_factor < 1:
        return 1
    if freq_factor > 1:
        return -1
    return 0


def get_frequency_factor(freq_one: str, freq_two: str) -> float:
    """Get relative factor between the two frequencies.

    Parameters
    ----------
    freq_one : str
        String representation of the first frequency.
    freq_two : str
        String representation of the second frequency.

    Returns
    -------
    float
        Ratio of `freq_one` to `freq_two`.

    Raises
    ------
    ValueError
        If the frequency factor cannot be determined. This can be the case if the frequencies do not represent a fixed
        time delta and are not directly comparable (e.g., because they have the same unit)
        E.g., a month does not represent a fixed time delta. Thus, 1D and 1M are not comparable. However, 1M and 2M are
        comparable since they have the same unit.
    """
    if freq_one == freq_two:
        return 1

    offset_one = to_offset(freq_one)
    offset_two = to_offset(freq_two)
    if offset_one.n < 0 or offset_two.n < 0:
        # Would be possible to implement, but we should never need negative frequencies, so it seems reasonable to
        # fail gracefully rather than to open ourselves to potential unexpected corner cases.
        raise NotImplementedError('Cannot compare negative frequencies.')
    # avoid division by zero errors
    if offset_one.n == offset_two.n == 0:
        return 1
    if offset_two.n == 0:
        return np.inf
    if offset_one.name == offset_two.name:
        return offset_one.n / offset_two.n

    # some simple hard-coded cases
    factor = None
    regex_month_or_day = '-(JAN|FEB|MAR|APR|MAY|JUN|JUL|AUG|SEP|OCT|NOV|DEC|MON|TUE|WED|THU|FRI|SAT|SUN)$'
    for i, (one, two) in enumerate([(offset_one, offset_two), (offset_two, offset_one)]):
        # the offset anchor is irrelevant for the ratio between the frequencies, so we remove it from the string
        name_one = re.sub(regex_month_or_day, '', one.name)
        name_two = re.sub(regex_month_or_day, '', two.name)
        if (name_one in ['A', 'Y'] and name_two == 'M') or (name_one in ['AS', 'YS'] and name_two == 'MS'):
            factor = 12 * one.n / two.n
        if (name_one in ['A', 'Y'] and name_two == 'Q') or (name_one in ['AS', 'YS'] and name_two == 'QS'):
            factor = 4 * one.n / two.n
        if (name_one == 'Q' and name_two == 'M') or (name_one == 'QS' and name_two == 'MS'):
            factor = 3 * one.n / two.n
        if name_one == 'W' and name_two == 'D':
            factor = 7 * one.n / two.n

        if factor is not None:
            if i == 1:
                return 1 / factor  # `one` was `offset_two`, `two` was `offset_one`
            return factor

    # If all other checks didn't match, we try to convert the frequencies to timedeltas. However, we first need to avoid
    # two cases: (1) pd.to_timedelta currently interprets 'M' as minutes, while it means months in to_offset.
    # (2) Using 'M', 'Y', and 'y' in pd.to_timedelta is deprecated and won't work in the future, so we don't allow it.
    if any(re.sub(regex_month_or_day, '', offset.name) in ['M', 'Y', 'A', 'y'] for offset in [offset_one, offset_two]):
        raise ValueError(f'Frequencies {freq_one} and/or {freq_two} are not comparable.')
    try:
        factor = pd.to_timedelta(freq_one) / pd.to_timedelta(freq_two)
    except ValueError as err:
        raise ValueError(f'Frequencies {freq_one} and/or {freq_two} are not comparable.') from err
    return factor



from pathlib import Path

from neuralhydrology.evaluation.tester import BaseTester, RegressionTester, UncertaintyTester
from neuralhydrology.utils.config import Config


def get_tester(cfg: Config, run_dir: Path, period: str, init_model: bool) -> BaseTester:
    """Get specific tester class objects depending on the model (head) type.
    
    Parameters
    ----------
    cfg : Config
        The run configuration.
    run_dir : Path
        Path to the run directory.
    period : {'train', 'validation', 'test'}
        The period to evaluate.
    init_model : bool
        If True, the model weights will be initialized with the checkpoint from the last available epoch in `run_dir`.

    Returns
    -------
    BaseTester
        `RegressionTester` if the model head is 'regression'. `UncertaintyTester` if the model head is one of 
        {'gmm', 'cmal', 'umal'} or if the evaluation is run in MC-Dropout mode.
    """
    if cfg.mc_dropout or cfg.head.lower() in ["gmm", "cmal", "umal"]:
        Tester = UncertaintyTester
    # MC-LSTM is a special case, where the head returns an empty string but the model is trained as regression model.
    elif cfg.head.lower() in ["regression", ""]:
        Tester = RegressionTester
    else:
        NotImplementedError(f"No evaluation method implemented for {cfg.head} head")

    return Tester(cfg=cfg, run_dir=run_dir, period=period, init_model=init_model)



"""Utility script to generate config files from a base config and a defined set of variations"""
import itertools
from pathlib import Path
from typing import Dict

from neuralhydrology.utils.config import Config


def create_config_files(base_config_path: Path, modify_dict: Dict[str, list], output_dir: Path):
    """Create configs, given a base config and a dictionary of parameters to modify.
    
    This function will create one config file for each combination of parameters defined in the modify_dict.
    
    Parameters
    ----------
    base_config_path : Path
        Path to a base config file (.yml)
    modify_dict : dict
        Dictionary, mapping from parameter names to lists of possible parameter values.
    output_dir : Path 
        Path to a folder where the generated configs will be stored
    """
    if not output_dir.is_dir():
        output_dir.mkdir(parents=True)

    # load base config as dictionary
    base_config = Config(base_config_path)
    experiment_name = base_config.experiment_name
    option_names = list(modify_dict.keys())

    # iterate over each possible combination of hyper parameters
    for i, options in enumerate(itertools.product(*[val for val in modify_dict.values()])):

        base_config.update_config(dict(zip(option_names, options)))

        # create a unique run name
        name = experiment_name
        for key, val in zip(option_names, options):
            name += f"_{key}{val}"
        base_config.update_config({"experiment_name": name})

        base_config.dump_config(output_dir, f"config_{i+1}.yml")

    print(f"Finished. Configs are stored in {output_dir}")



from typing import Sequence

import numpy as np


class RatingCurve(object):
    """Class to estimate rating curves from stage and discharge data and to convert stage to discharge and vice-versa.
            
    Parameters
    ----------
    stages : Sequence
        Stage data to estimate a rating curve 
    discharges : Sequence
        Discharge data to estimate a rating curve 
    move_stages_to_zero : bool, optional
        Option to account for any offset in the stage data. This will automatically set the minimum measured stage to 
        zero.
    """

    def __init__(self, stages: Sequence[float], discharges: Sequence[float], move_stages_to_zero: bool = True):

        # Validate input.
        if len(stages) != len(discharges):
            raise ValueError("The sequence 'stages' and 'discharges' must have the same length")

        self.stages = np.array(stages, dtype=np.float32)
        self.discharges = np.array(discharges, dtype=np.float32)

        # In the interpolation, x-axis is stage and y-axis is discharge.
        if move_stages_to_zero:
            use_stages = self.stages - np.min(self.stages)
            self.zero_stages = np.min(self.stages)
        else:
            use_stages = self.stages
            self.zero_stages = False

        self.coeffs = np.polyfit(use_stages, self.discharges, deg=2)

    def stage_to_discharge(self, stage: np.ndarray) -> np.ndarray:
        """Convert stage to discharge.
            
        Parameters
        ----------
        stage : np.ndarray
            Stage data to convert to discharge 
    
        Returns
        -------
        np.ndarray
            Estimated discharge 
        """

        if self.zero_stages:
            stage -= self.zero_stages

        return self.coeffs[2] + stage * self.coeffs[1] + stage**2 * self.coeffs[0]

    def discharge_to_stage(self, discharge: np.ndarray) -> np.ndarray:
        """Convert discharge to stage.
            
        Parameters
        ----------
        discharge : np.ndarray
            Discharge data to convert to stage 
  
        Returns
        -------
        np.ndarray
            Estimated stage 
        """

        # init storage
        stage = np.full(discharge.shape, np.nan)
        solutions = np.full([discharge.shape[0], 2], np.nan)

        centered_bias = self.coeffs[2] - discharge
        radicand = self.coeffs[1]**2 - 4 * self.coeffs[0] * centered_bias
        radicand[np.isnan(radicand)] = -9999
        mask = radicand >= 0

        solutions[mask, 0] = (-self.coeffs[1] + np.sqrt(radicand[mask])) / (2 * self.coeffs[0])
        solutions[mask, 1] = (-self.coeffs[1] - np.sqrt(radicand[mask])) / (2 * self.coeffs[0])

        best_idx = np.argmin(np.abs(solutions[mask] - np.expand_dims(discharge[mask], -1)), axis=1)

        stage[mask] = solutions[mask, best_idx]

        if self.zero_stages:
            stage += self.zero_stages

        return stage



import torch
from typing import Dict, Union
from neuralhydrology.modelzoo.baseconceptualmodel import BaseConceptualModel
from neuralhydrology.utils.config import Config

class SHM(BaseConceptualModel):
    """Modified version of SHM [#]_ hydrological model with dynamic parameterization.
    
    The SHM receives the dynamic parameterization given by a deep learning model. This class has two properties which define the initial 
    conditions of the internal states of the model (buckets) and the ranges in which the model parameters 
    are allowed to vary during optimization.

    Parameters
    ----------
    cfg : Config
        The run configuration.

    References
    ----------
    .. [#] Ehret, U., van Pruijssen, R., Bortoli, M., Loritz, R., Azmi, E., and Zehe, E.: Adaptive clustering: reducing
    the computational costs of distributed (hydrological) modelling by exploiting time-variable similarity among model
    elements, Hydrology and Earth System Sciences, 24, 4389–4411, https://doi.org/10.5194/hess-24-4389-2020, 2020.

    """
    def __init__(self, cfg: Config):
        super(SHM, self).__init__(cfg=cfg)

    def forward(self, x_conceptual: torch.Tensor, lstm_out: torch.Tensor) -> Dict[str, Union[torch.Tensor, Dict[str, torch.Tensor]]]:
        """Perform a forward pass on the SHM model. In this forward pass, all elements of the batch are processed in  parallel

        Parameters
        ----------
        x_conceptual: torch.Tensor
            Tensor of size [batch_size, time_steps, n_inputs]. The batch_size is associated with a certain basin and a
            certain prediction period. The time_steps refer to the number of time steps (e.g. days) that our conceptual
            model is going to be run for. The n_inputs refer to the dynamic forcings used to run the conceptual model
            (e.g. Precipitation, Temperature...)

        lstm_out: torch.Tensor
            Tensor of size [batch_size, time_steps, n_parameters]. The tensor comes from the data-driven model  and will
            be used to obtained the dynamic parameterization of the conceptual model

        Returns
        -------
        Dict[str, Union[torch.Tensor, Dict[str, torch.Tensor]]]
            - y_hat: torch.Tensor
                Simulated outflow
            - parameters: Dict[str, torch.Tensor]
                Dynamic parameterization of the conceptual model
            - internal_states: Dict[str, torch.Tensor]]
                Time-evolution of the internal states of the conceptual model

        """
        # get model parameters
        parameters = self._get_dynamic_parameters_conceptual(lstm_out=lstm_out)

        # initialize structures to store the information
        states, out = self._initialize_information(conceptual_inputs=x_conceptual)

        # initialize constants
        zero = torch.tensor(0.0, dtype=torch.float32, device=x_conceptual.device)
        one = torch.tensor(1.0, dtype=torch.float32, device=x_conceptual.device)
        klu = torch.tensor(0.90, requires_grad=False, dtype=torch.float32, device=x_conceptual.device)  # land use correction factor [-]

        # auxiliary vectors to accelerate the execution of the hydrological model
        t_mean = (x_conceptual[:, :, 2] + x_conceptual[:, :, 3]) / 2
        temp_mask = t_mean < 0
        snow_melt = t_mean * parameters['dd']
        snow_melt[temp_mask] = zero
        # liquid precipitation:
        liquid_p = x_conceptual[:, :, 0].clone()
        liquid_p[temp_mask] = zero
        # solid precipitation (snow):
        snow = x_conceptual[:, :, 0].clone()
        snow[~temp_mask] = zero
        # permanent wilting point use in ET:
        pwp = torch.tensor(0.8, dtype=torch.float32, device=x_conceptual.device) * parameters['sumax']

        # Storages
        ss = torch.tensor(self.initial_states['ss'], dtype=torch.float32, device=x_conceptual.device).repeat(x_conceptual.shape[0])
        sf = torch.tensor(self.initial_states['sf'], dtype=torch.float32, device=x_conceptual.device).repeat(x_conceptual.shape[0])
        su = torch.tensor(self.initial_states['su'], dtype=torch.float32, device=x_conceptual.device).repeat(x_conceptual.shape[0])
        si = torch.tensor(self.initial_states['si'], dtype=torch.float32, device=x_conceptual.device).repeat(x_conceptual.shape[0])
        sb = torch.tensor(self.initial_states['ss'], dtype=torch.float32, device=x_conceptual.device).repeat(x_conceptual.shape[0])

        # run hydrological model for each time step
        for j in range(x_conceptual.shape[1]):
            # Snow module --------------------------
            qs_out = torch.minimum(ss, snow_melt[:, j])
            ss = ss - qs_out + snow[:, j]
            qsp_out = qs_out + liquid_p[:, j]

            # Split snowmelt+rainfall into inflow to fastflow reservoir and unsaturated reservoir ------
            qf_in = torch.maximum(zero, qsp_out - parameters['f_thr'][:, j])
            qu_in = torch.minimum(qsp_out, parameters['f_thr'][:, j])

            # Fastflow module ----------------------
            sf = sf + qf_in
            qf_out = sf / parameters['kf'][:, j]
            sf = sf - qf_out

            # Unsaturated zone----------------------
            psi = (su / parameters['sumax'][:, j]) ** parameters['beta'][:, j]  # [-]
            su_temp = su + qu_in * (1 - psi)
            su = torch.minimum(su_temp, parameters['sumax'][:, j])
            qu_out = qu_in * psi + torch.maximum(zero, su_temp - parameters['sumax'][:, j])  # [mm]
            # Evapotranspiration -------------------
            ktetha = su / parameters['sumax'][:, j]
            et_mask = su <= pwp[:, j]
            ktetha[~et_mask] = one
            ret = x_conceptual[:, j, 1] * klu * ktetha  # [mm]
            su = torch.maximum(zero, su - ret)  # [mm]

            # Interflow reservoir ------------------
            qi_in = qu_out * parameters['perc'][:, j] # [mm]
            si = si + qi_in  # [mm]
            qi_out = si / parameters['ki'][:, j] # [mm]
            si = si - qi_out  # [mm]

            # Baseflow reservoir -------------------
            qb_in = qu_out * (1.0 - parameters['perc'][:, j])  # [mm]
            sb = sb + qb_in  # [mm]
            qb_out = sb / parameters['kb'][:, j]  # [mm]
            sb = sb - qb_out

            # Store time evolution of the internal states
            states['ss'][:, j] = ss
            states['sf'][:, j] = sf
            states['su'][:, j] = su
            states['sb'][:, j] = sb

            # total outflow
            out[:, j, 0] = qf_out + qi_out + qb_out  # [mm]

        return {'y_hat': out, 'parameters': parameters, 'internal_states': states}

    @property
    def initial_states(self):
        return {'ss': 0.0,
                'sf': 1.0,
                'su': 5.0,
                'si': 10.0,
                'sb': 15.0}

    @property
    def parameter_ranges(self):
        return {'dd': [0.0, 10.0],
                'f_thr': [10.0, 60.0],
                'sumax': [20.0, 700.0],
                'beta': [1.0, 6.0],
                'perc': [0.0, 1.0],
                'kf': [1.0, 20.0],
                'ki': [1.0, 100.0],
                'kb': [10.0, 1000.0]
                }

